//---------------------------------------------------------------------------
#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
//---------------------------------------------------------------------------
#define BLE_Name     "ESP32_BLE"
#define UUID_SERVICE "a8df5590-9515-11ea-bb37-0242ac130002"
#define UUID_RW      "a8df5798-9515-11ea-bb37-0242ac130002"
//---------------------------------------------------------------------------
BLEServer*			pBLEServer;
BLEService*			pBLEService;
BLECharacteristic*	pBLEReadWrite;
//---------------------------------------------------------------------------
String BLEReceive = "";
class BLEReadWrite_Callback: public BLECharacteristicCallbacks
{
	void onWrite(BLECharacteristic* pBLECharacteristic)
	{
		std::string strValue = pBLECharacteristic->getValue();
		for (int nIndex = 0; nIndex < strValue.length(); nIndex++) BLEReceive += String(strValue[nIndex]);
	}
};
//---------------------------------------------------------------------------
boolean BLE_Init()
{
	//初始化BLE设备
	BLEDevice::init(BLE_Name);

	//建立BLE服务器
	pBLEServer = BLEDevice::createServer();
	if (pBLEServer == nullptr) return false;

	//建立BLE服务
	pBLEService = pBLEServer->createService(UUID_SERVICE);
	if (pBLEService == nullptr) return false;

	//建立BLE接收和发送
	pBLEReadWrite = pBLEService->createCharacteristic(UUID_RW, BLECharacteristic::PROPERTY_WRITE | BLECharacteristic::PROPERTY_READ | BLECharacteristic::PROPERTY_NOTIFY);
	if (pBLEReadWrite == nullptr) return false;
	pBLEReadWrite->setCallbacks(new BLEReadWrite_Callback());

	//启动BLE服务
	pBLEService->start();

	//启动BLE服务广播
//	BLEAdvertising* pAdvertising = pBLEServer->getAdvertising();  //this still is working for backward compatibility
	BLEAdvertising* pAdvertising = BLEDevice::getAdvertising();
	pAdvertising->addServiceUUID(UUID_SERVICE);
	pAdvertising->setScanResponse(true);
	pAdvertising->setMinPreferred(0x06);  //functions that help with iPhone connections issue
	pAdvertising->setMinPreferred(0x12);
	BLEDevice::startAdvertising();

	Serial.print("BLE Name: ");			Serial.println(BLE_Name);
	Serial.print("UUID(Service): ");	Serial.println(UUID_SERVICE);
	Serial.print("UUID(RW): ");			Serial.println(UUID_RW);
	return true;
}
//---------------------------------------------------------------------------
void BLE_Write(const String strValue)
{
	std::string strSend = "";
	strSend.append(strValue.c_str());
	pBLEReadWrite->setValue(strSend);
	pBLEReadWrite->notify();
}
//---------------------------------------------------------------------------
String BLE_Receive()
{
	String strReceive = "";

	int nIndex = BLEReceive.indexOf("\n");
	if (nIndex >= 0)
	{
		strReceive = BLEReceive.substring(0, nIndex);
		BLEReceive = BLEReceive.substring(nIndex + 1, BLEReceive.length());
	}

	return strReceive;
}
//---------------------------------------------------------------------------
void BLE_Send(String strSend)
{
	if (pBLEServer->getConnectedCount() > 0)
	{
		strSend += "\n";
		BLE_Write(strSend);
	}
}
//---------------------------------------------------------------------------
