//---------------------------------------------------------------------------
#include "Global.h"
//---------------------------------------------------------------------------
//任务
void Task_AutoPatrol	(void* pvParameters);	//自动巡逻
void Task_SR04			(void* pvParameters);	//超声波测距
void Task_Ctrl			(void* pvParameters);	//接收控制内容
void Task_Feedback		(void* pvParameters);	//反馈数据
void Task_Button		(void* pvParameters);	//检测按钮
void Task_Core0			(void* pvParameters);	//其他
//---------------------------------------------------------------------------
void setup()
{
	//初始化串口
	Serial.begin(115200);
	while (! Serial);
	Serial.println("\nprogram start");

	//初始化I2C
	Serial.println("Wire ...");
	if (Wire.begin())
	{
		Wire.setClock(400000);
		Serial.println("Wire ok");
	}
	else
	{
		Serial.println("Wire failure");
	}

	//初始化显示屏
	Display_Init();
	Display_Label(0, "initialization ...");

	//初始化急停引脚
	pinMode(EmeStop_Pin, INPUT_PULLUP);

	//初始化电压ADC引脚
	pinMode(Power_Pin, INPUT);

	//初始化超声波测距引脚
	pinMode(US_Trig_Pin, OUTPUT);
	pinMode(US_Echo_Pin, INPUT);

	//初始化左轮引脚
	pinMode(L_IN1_Pin, OUTPUT);
	pinMode(L_IN2_Pin, OUTPUT);
	pinMode(L_PWM_Pin, OUTPUT);
	pinMode(L_ENA_Pin, INPUT);
	pinMode(L_ENB_Pin, INPUT);

	//初始化右轮引脚
	pinMode(R_IN1_Pin, OUTPUT);
	pinMode(R_IN2_Pin, OUTPUT);
	pinMode(R_PWM_Pin, OUTPUT);
	pinMode(R_ENA_Pin, INPUT);
	pinMode(R_ENB_Pin, INPUT);

	//初始化编码器中断（4倍频）
	attachInterrupt(digitalPinToInterrupt(L_ENA_Pin), L_ENA_Pin_Interrupt, CHANGE);
	attachInterrupt(digitalPinToInterrupt(L_ENB_Pin), L_ENB_Pin_Interrupt, CHANGE);
	attachInterrupt(digitalPinToInterrupt(R_ENA_Pin), R_ENA_Pin_Interrupt, CHANGE);
	attachInterrupt(digitalPinToInterrupt(R_ENB_Pin), R_ENB_Pin_Interrupt, CHANGE);

	//初始化电机
	PWM_Ctrl(0, 0);

	//初始化参数
	Display_Label(1, "EEPROM ...");
	if (EEPROM.begin(0x100 * sizeof(float) + 2100))
	{
		ReadAllParams();
		Display_Label(1, "EEPROM ok");
	}
	else
	{
		Display_Label(1, "EEPROM failure");
	}

	//初始化姿态模块
	Display_Label(2, "MPU6050 ...");
	Serial2.begin(115200, SERIAL_8N1, MPU6050_TX_Pin, MPU6050_RX_Pin);
	if (MPU6050.begin(&Serial2)) Display_Label(2, "MPU6050 ok"); else Display_Label(2, "MPU6050 failure");

	//初始化超声波测距模块
	Display_Label(3, "SR04 ...");
	if (Ultrasonic_Init()) Display_Label(3, "SR04 ok"); else Display_Label(3, "SR04 failure");

	//初始化是否启用数据通讯功能（上电的时候，如果不按住急停按钮，则启用数据通讯功能）
	USE_Communicate = (digitalRead(EmeStop_Pin) == HIGH);

	//初始化WIFI
	#ifdef USE_WIFI
		if (USE_Communicate)
		{
			Display_Label(4, "WIFI ...");
			if (WIFI_Init())
			{
				//初始化TCP Server
				Display_Label(4, "TCP Server ...");
				if (TCP_Server_Init()) Display_Label(4, "TCP " + WiFi.localIP().toString() + ":" + String(TCP_Port)); else Display_Label(4, "TCP Server failure");
			}
			else
			{
				Display_Label(4, "WIFI failure");
			}
		}
		else
		{
			Display_Label(4, "WIFI ignore");
		}
	#endif

	//初始化BLE
	#ifdef USE_BLE
		if (USE_Communicate)
		{
			Display_Label(4, "BLE ...");
			if (BLE_Init()) Display_Label(4, "BLE ok (" + String(BLE_Name) + ")"); else Display_Label(4, "BLE failure");
		}
		else
		{
			Display_Label(4, "BLE ignore");
		}
	#endif

	//初始化BT_SERIAL
	#ifdef USE_BT_SERIAL
		if (USE_Communicate)
		{
			Display_Label(4, "BT SERIAL ...");
			if (BT_SERIAL_Init()) Display_Label(4, "BT SERIAL ok (" + String(BT_SERIAL_Name) + ")"); else Display_Label(4, "BT SERIAL failure");
		}
		else
		{
			Display_Label(4, "BT SERIAL ignore");
		}
	#endif

	//初始化WIRELESS_SERIAL
	#ifdef USE_WIRELESS_SERIAL
		if (USE_Communicate)
		{
			Display_Label(5, "WIRELESS SERIAL ...");
			if (WIRELESS_SERIAL_Init(&Serial1, WIRELESS_SERIAL_Baud, Wireless_TX_Pin, Wireless_RX_Pin)) Display_Label(5, "WIRELESS SERIAL ok"); else Display_Label(5, "WIRELESS SERIAL failure");
		}
		else
		{
			Display_Label(5, "WIRELESS SERIAL ignore");
		}
	#endif

	//初始化任务
	Display_Label(6, "task ...");
	bool boolTask = true;
	boolTask = (boolTask && (xTaskCreatePinnedToCore(Task_AutoPatrol,	"Task_AutoPatrol",	2000, NULL, 1, NULL, 0) == pdPASS));
	boolTask = (boolTask && (xTaskCreatePinnedToCore(Task_SR04,			"Task_SR04",		2000, NULL, 1, NULL, 0) == pdPASS));
	boolTask = (boolTask && (xTaskCreatePinnedToCore(Task_Ctrl,			"Task_Ctrl",		2000, NULL, 1, NULL, 0) == pdPASS));
	boolTask = (boolTask && (xTaskCreatePinnedToCore(Task_Feedback,		"Task_Feedback",	2000, NULL, 1, NULL, 0) == pdPASS));
	boolTask = (boolTask && (xTaskCreatePinnedToCore(Task_Button,		"Task_Button",		2000, NULL, 1, NULL, 0) == pdPASS));
	boolTask = (boolTask && (xTaskCreatePinnedToCore(Task_Core0,		"Task_Core0",		3000, NULL, 1, NULL, 0) == pdPASS));
	if (boolTask) Display_Label(6, "task ok"); else Display_Label(6, "task failure");

	//初始化完成
	Display_Label(0, "initialization ok");
}
//---------------------------------------------------------------------------
void loop()
{
	static unsigned long nLastTime = millis();
	unsigned long now = millis();
	if (now - nLastTime >= PIDTime)
	{
		nLastTime = now;

		//获取姿态
		GetAngle();

		//PWM计算
		PWM_Calc();

		//PWM输出
		PWM_Ctrl(L_PWM, R_PWM);
	}
}
//---------------------------------------------------------------------------
void Task_AutoPatrol(void* pvParameters)
{
	while (true)
	{
		//自动巡逻
		doAutoPatrol();
		vTaskDelay(1);
	}
}
//---------------------------------------------------------------------------
void Task_SR04(void* pvParameters)
{
	while (true)
	{
		//超声波测距
		Ultrasonic_Ranging();
		vTaskDelay(1);
	}
}
//---------------------------------------------------------------------------
void Task_Ctrl(void* pvParameters)
{
	while (true)
	{
		//接收控制内容
		Ctrl_Receive();
		vTaskDelay(1);
	}
}
//---------------------------------------------------------------------------
void Task_Feedback(void* pvParameters)
{
	while (true)
	{
		//反馈数据
		FeedbackData();
		vTaskDelay(Feedback_Interval);
	}
}
//---------------------------------------------------------------------------
void Task_Button(void* pvParameters)
{
	while (true)
	{
		//检测按钮
		switch (ButtonClick())
		{
			//单击
			case 1:
				break;

			//双击
			case 2:
				//切换显示内容
				DisplayType++;
				break;

			//三击
			case 3:
				break;
		}

		vTaskDelay(1);
	}
}
//---------------------------------------------------------------------------
void Task_Core0(void* pvParameters)
{
	unsigned long nStartTime = millis();

	while (true)
	{
		//获取供电电量百分比
		GetPowerPercent();

		//显示内容
		static bool boolDisplayContent = false;
		if (boolDisplayContent)
		{
			Display_Content();
		}
		else
		{
			boolDisplayContent = (millis() - nStartTime >= 5000);
			if (boolDisplayContent) Display_Init();
		}

		//默认看门狗CONFIG_TASK_WDT_TIMEOUT_S=5秒，如果程序运行期间没有任何空闲时间，会导致系统无法自动喂狗，或者索性禁用看门狗disableCore0WDT();
		vTaskDelay(1);
	}
}
//---------------------------------------------------------------------------
