//---------------------------------------------------------------------------
/*
角度：右倾为负，左倾为正
     前倾为负，后仰为正
     右转为负，左转为正
左轮：向前为正，向后为反
右轮：向前为反，向后为正

反馈编码：
00：实际PID计算周期
01：X轴角度
02：Y轴角度
03：Z轴角度
04：X轴角速度
05：Y轴角速度
06：Z轴角速度
07：X轴加速度
08：Y轴加速度
09：Z轴加速度
10：超声波测距
11：PWM
12：转向PWM
13：左轮PWM
14：右轮PWM
15：左轮编码器计数
16：右轮编码器计数
17：供电电量百分比
21：平衡角度
22：移动速度
23：转向角速度
28：是否自动巡逻
29：是否反馈数据
31：角度环kp
32：角度环ki
33：角度环kd
34：角度环kf
35：角度环km
36：角度环kn
41：速度环kp
42：速度环ki
43：速度环kd
44：速度环kf
45：速度环km
46：速度环kn
51：转向环kp
52：转向环ki
53：转向环kd
54：转向环kf
55：转向环km
56：转向环kn
81：急停状态
91：WIRELESS_SERIAL波特率
92：反馈数据间隔时间
93：供电电池数量

接收编码：
A0：停止
A1：平衡角度
A2：移动速度
A3：转向角速度
A8：是否自动巡逻
A9：是否反馈数据
B1：角度环kp
B2：角度环ki
B3：角度环kd
B4：角度环kf
B5：角度环km
B6：角度环kn
C1：速度环kp
C2：速度环ki
C3：速度环kd
C4：速度环kf
C5：速度环km
C6：速度环kn
D1：转向环kp
D2：转向环ki
D3：转向环kd
D4：转向环kf
D5：转向环km
D6：转向环kn
F1：WIRELESS_SERIAL波特率
F2：反馈数据间隔时间
F3：供电电池数量
*/
//---------------------------------------------------------------------------
//#define USE_WIFI
//#define USE_BLE
#define USE_BT_SERIAL
#define USE_WIRELESS_SERIAL
//---------------------------------------------------------------------------
#include <U8g2lib.h>
#include <EEPROM.h>
#include "MPU6050.h"
//---------------------------------------------------------------------------
#ifdef USE_WIFI
	#include "WIFI.h"
#endif
//---------------------------------------------------------------------------
#ifdef USE_BLE
	#include "BLE.h"
#endif
//---------------------------------------------------------------------------
#ifdef USE_BT_SERIAL
	#include "BT_SERIAL.h"
#endif
//---------------------------------------------------------------------------
#ifdef USE_WIRELESS_SERIAL
	#include "WIRELESS_SERIAL.h"
#endif
//---------------------------------------------------------------------------
//急停引脚
#define		EmeStop_Pin		23		//急停（LOW）
//---------------------------------------------------------------------------
//电压ADC引脚
#define		Power_Pin		39		//电位器抽头（VN引脚）
//---------------------------------------------------------------------------
//显示屏引脚
#define		Display_SCL_Pin	22		//I2C SCL
#define		Display_SDA_Pin	21		//I2C SDA
//---------------------------------------------------------------------------
//MPU6050引脚
#define		MPU6050_RX_Pin	17		//模块RX（TX2引脚）
#define		MPU6050_TX_Pin	16		//模块TX（RX2引脚）
//---------------------------------------------------------------------------
//超声波测距引脚
#define		US_Trig_Pin		5		//触发
#define		US_Echo_Pin		19		//反馈
//---------------------------------------------------------------------------
//无线透传引脚
#define		Wireless_RX_Pin	15		//模块RX
#define		Wireless_TX_Pin	4		//模块TX
#define		Wireless_CS_Pin	0		//模块片选（低电平唤醒，高电平睡眠）
//---------------------------------------------------------------------------
//左轮引脚
#define		L_IN1_Pin		13		//相位1（AIN1）
#define		L_IN2_Pin		12		//相位2（AIN2）
#define		L_PWM_Pin		14		//PWM
#define		L_PWM_Chl		1		//PWM通道号
#define		L_ENA_Pin		27		//编码器A相
#define		L_ENB_Pin		26		//编码器B相
//---------------------------------------------------------------------------
//右轮引脚
#define		R_IN1_Pin		25		//相位1（BIN1）
#define		R_IN2_Pin		33		//相位2（BIN2）
#define		R_PWM_Pin		32		//PWM
#define		R_PWM_Chl		2		//PWM通道号
#define		R_ENA_Pin		35		//编码器A相
#define		R_ENB_Pin		34		//编码器B相
//---------------------------------------------------------------------------
#define		PIDTime			10		//PID计算周期（毫秒），最好是和姿态计算周期一致，否则会引起不稳
#define		PWMMax			255		//PWM最大值
#define		ENCMax			100		//PWM最大值情况下，一个PID计算周期内的编码器最大值
#define		MoveSpeedMax	160		//移动速度最大值
#define		TurnSpeedMax	180		//转向角速度最大值
//---------------------------------------------------------------------------
U8G2_SH1106_128X64_NONAME_F_HW_I2C u8g2(U8G2_R0, U8X8_PIN_NONE);	//u8g2显示屏
int			Power_Percent;			//供电电量百分比
TMPU6050	MPU6050;				//MPU6050姿态
int			SR04;					//SR04超声波测距
boolean		USE_Communicate;		//是否启用数据通讯功能
//---------------------------------------------------------------------------
//左轮编码器计数
int			L_ENC			= 0;	//小计
int			L_ENC_Count		= 0;	//总计
int			L_ENC_Cycle		= 0;	//一个PID计算周期
portMUX_TYPE L_ENC_MUX		= portMUX_INITIALIZER_UNLOCKED;
//---------------------------------------------------------------------------
//右轮编码器计数
int			R_ENC			= 0;	//小计
int			R_ENC_Count		= 0;	//总计
int			R_ENC_Cycle		= 0;	//一个PID计算周期
portMUX_TYPE R_ENC_MUX		= portMUX_INITIALIZER_UNLOCKED;
//---------------------------------------------------------------------------
float		MoveSpeed		= 0;	//移动速度（一个PID计算周期内左右编码器计数之和）
									/*
										电机减速比：1:34
										霍尔编码器：12CPR，4倍频，左右2个
										轮子转一圈：1*34*12*4*2=3264编码器计数
										轮子直径：75mm
										轮子周长：75*3.1415926=235.62mm
										PID计算周期：10ms
										移动速度：1/3264*235.62/10*1000=7.21875mm/s
									*/
float		TurnSpeed		= 0;	//转向角速度（°/s）
int			AutoPatrol		= 0;	//是否自动巡逻
int			Feedback		= 0;	//是否反馈数据
//---------------------------------------------------------------------------
int			PIDTimeLoop;			//实际PID计算周期（毫秒）
float		Balance;				//平衡角度（°）
//---------------------------------------------------------------------------
int			PWM;					//PWM
int			PWMTurn;				//转向PWM
int			L_PWM;					//左轮PWM
int			R_PWM;					//右轮PWM
//---------------------------------------------------------------------------
float		kap, ksp, ktp;			//比例系数    13.72  26.60   9.50
float		kai, ksi, kti;			//积分系数     0.00   0.13   0.47
float		kad, ksd, ktd;			//微分系数     0.44   0.00   0.00
float		kaf, ksf, ktf;			//低通滤波     1.00   0.10   1.00
float		kam, ksm, ktm;			//整体比例     0.80   0.90   1.00（静止状态，12V）
float		kan, ksn, ktn;			//整体比例     0.70   0.85   1.00（运动状态，12V）
//---------------------------------------------------------------------------
int			DisplayType		= 1;	//显示内容，0：关闭显示；1-n：显示内容
//---------------------------------------------------------------------------
int			EmergencyStopStatus;	//急停状态
//---------------------------------------------------------------------------
int			WIRELESS_SERIAL_Baud;	//WIRELESS_SERIAL波特率
int			Feedback_Interval;		//反馈数据间隔时间
int			Power_BatteryNum;		//供电电池数量
//---------------------------------------------------------------------------
