//---------------------------------------------------------------------------
#include "Define.h"
//---------------------------------------------------------------------------
//输出模拟量
void analogWrite(const byte pin, const int value)
{
	//左轮
	static bool L_SetChannel = false;
	if (pin == L_PWM_Pin)
	{
		if (! L_SetChannel)
		{
			ledcSetup(L_PWM_Chl, 10000, 8);
			ledcAttachPin(pin, L_PWM_Chl);
			L_SetChannel = true;
		}

		ledcWrite(L_PWM_Chl, value);
	}

	//右轮
	static bool R_SetChannel = false;
	if (pin == R_PWM_Pin)
	{
		if (! R_SetChannel)
		{
			ledcSetup(R_PWM_Chl, 10000, 8);
			ledcAttachPin(pin, R_PWM_Chl);
			R_SetChannel = true;
		}

		ledcWrite(R_PWM_Chl, value);
	}
}
//---------------------------------------------------------------------------
//读取参数
float ReadParam(const String strCtrlType)
{
	char c[4];
	strCtrlType.toCharArray(c, sizeof(c));
	uint16_t nCtrlType;
	sscanf(c, "%x", &nCtrlType);

	float fResult = 0;
	if (nCtrlType < 0x100) fResult = EEPROM.readFloat(nCtrlType * sizeof(float) + 2100);  //0-2047被Blinker占用
	if (isnan(fResult)) fResult = 0;
	return fResult;
}
//---------------------------------------------------------------------------
//读取全部参数
void ReadAllParams()
{
	EmergencyStopStatus		= ReadParam("81");	Serial.print("EmergencyStopStatus: ");	Serial.println(EmergencyStopStatus);	//急停状态
	Balance					= ReadParam("A1");	Serial.print("Balance: ");				Serial.println(Balance);				//平衡角度
	kap						= ReadParam("B1");	Serial.print("kap: ");					Serial.println(kap);					//角度环kp
	kai						= ReadParam("B2");	Serial.print("kai: ");					Serial.println(kai);					//角度环ki
	kad						= ReadParam("B3");	Serial.print("kad: ");					Serial.println(kad);					//角度环kd
	kaf						= ReadParam("B4");	Serial.print("kaf: ");					Serial.println(kaf);					//角度环kf
	kam						= ReadParam("B5");	Serial.print("kam: ");					Serial.println(kam);					//角度环km
	kan						= ReadParam("B6");	Serial.print("kan: ");					Serial.println(kan);					//角度环kn
	ksp						= ReadParam("C1");	Serial.print("ksp: ");					Serial.println(ksp);					//速度环kp
	ksi						= ReadParam("C2");	Serial.print("ksi: ");					Serial.println(ksi);					//速度环ki
	ksd						= ReadParam("C3");	Serial.print("ksd: ");					Serial.println(ksd);					//速度环kd
	ksf						= ReadParam("C4");	Serial.print("ksf: ");					Serial.println(ksf);					//速度环kf
	ksm						= ReadParam("C5");	Serial.print("ksm: ");					Serial.println(ksm);					//速度环km
	ksn						= ReadParam("C6");	Serial.print("ksn: ");					Serial.println(ksn);					//速度环kn
	ktp						= ReadParam("D1");	Serial.print("ktp: ");					Serial.println(ktp);					//转向环kp
	kti						= ReadParam("D2");	Serial.print("kti: ");					Serial.println(kti);					//转向环ki
	ktd						= ReadParam("D3");	Serial.print("ktd: ");					Serial.println(ktd);					//转向环kd
	ktf						= ReadParam("D4");	Serial.print("ktf: ");					Serial.println(ktf);					//转向环kf
	ktm						= ReadParam("D5");	Serial.print("ktm: ");					Serial.println(ktm);					//转向环km
	ktn						= ReadParam("D6");	Serial.print("ktn: ");					Serial.println(ktn);					//转向环kn
	WIRELESS_SERIAL_Baud	= ReadParam("F1");	Serial.print("WIRELESS_SERIAL_Baud: ");	Serial.println(WIRELESS_SERIAL_Baud);	//WIRELESS_SERIAL波特率
	Feedback_Interval		= ReadParam("F2");	Serial.print("Feedback_Interval: ");	Serial.println(Feedback_Interval);		//反馈数据间隔时间
	Power_BatteryNum		= ReadParam("F3");	Serial.print("Power_BatteryNum: ");		Serial.println(Power_BatteryNum);		//供电电池数量
}
//---------------------------------------------------------------------------
//保存参数
void WriteParam(const String strCtrlType, const float fCtrlValue)
{
	char c[4];
	strCtrlType.toCharArray(c, sizeof(c));
	uint16_t nCtrlType;
	sscanf(c, "%x", &nCtrlType);

	if (nCtrlType < 0x100)
	{
		EEPROM.writeFloat(nCtrlType * sizeof(float) + 2100, fCtrlValue);  //0-2047被Blinker占用
		EEPROM.commit();
	}
}
//---------------------------------------------------------------------------
//获取移动速度
float GetMoveSpeed(bool boolForce = false)
{
	if (boolForce) return MoveSpeed;
	if (abs(MPU6050.Data.AngleY) > 45) return 0;
	return MoveSpeed;
}
//---------------------------------------------------------------------------
//获取转向角速度
float GetTurnSpeed(bool boolForce = false)
{
	if (boolForce) return TurnSpeed;
	if (abs(MPU6050.Data.AngleY) > 45) return 0;
	return TurnSpeed;
}
//---------------------------------------------------------------------------
//初始化显示屏
void Display_Init()
{
	u8g2.begin();
	u8g2.setDisplayRotation(U8G2_R2);
	u8g2.enableUTF8Print();
	u8g2.setFont(u8g2_font_ncenB08_tr);
    u8g2.setFontDirection(0);
	u8g2.setFontPosTop();
	u8g2.setFontMode(0);
	u8g2.setDrawColor(1);
	u8g2.clear();
}
//---------------------------------------------------------------------------
//显示标签
void Display_Label(const int nLabel, const String strCaption)
{
	static String strLabelCaption[10] = {"", "", "", "", "", "", "", "", "", ""};
	strLabelCaption[nLabel] = strCaption;

	u8g2.clearBuffer();
	for (int nIndex = 0; nIndex < 10; nIndex++) u8g2.drawUTF8(0, nIndex * 9, strLabelCaption[nIndex].c_str());
	u8g2.sendBuffer();

	Serial.print(nLabel);
	Serial.print(". ");
	Serial.println(strCaption);
}
//---------------------------------------------------------------------------
//显示内容
void Display_Content()
{
	static bool boolDisplay = true;
	String strLabelCaption[10] = {"", "", "", "", "", "", "", "", "", ""};

	if (DisplayType == 1)
	{
		strLabelCaption[0] = "Power: "		+ String(Power_Percent) + "%";										//供电电量百分比
		strLabelCaption[1] = "PID Time: "	+ String(PIDTimeLoop) + "ms";										//实际PID计算周期
		strLabelCaption[2] = "Angle: "		+ String(MPU6050.Data.AngleY);										//Y轴角度
		strLabelCaption[3] = "SR04: "		+ String(SR04) + "mm";												//超声波测距
		strLabelCaption[4] = "ENC: "		+ String(L_ENC_Count) + ", " + String(R_ENC_Count);					//左轮、右轮编码器计数
		strLabelCaption[5] = "PWM: "		+ String(L_PWM) + ", " + String(R_PWM);								//左轮、右轮PWM
		strLabelCaption[6] = "Speed: "		+ String(GetMoveSpeed(true)) + ", " + String(GetTurnSpeed(true));	//移动速度、转向角速度
	}
	else if (DisplayType == 2)
	{
		strLabelCaption[0] = "Balance: "	+ String(Balance);													//平衡角度
		strLabelCaption[1] = "Angle Y: "	+ String(MPU6050.Data.AngleY);										//Y轴角度
		strLabelCaption[2] = "Gyro  Y: "	+ String(MPU6050.Data.GyroY);										//Y轴角速度
		strLabelCaption[3] = "Acc   Y: "	+ String(MPU6050.Data.AccY);										//Y轴加速度
		strLabelCaption[4] = "Angle Z: "	+ String(MPU6050.Data.AngleZ);										//Z轴角度
		strLabelCaption[5] = "Gyro  Z: "	+ String(MPU6050.Data.GyroZ);										//Z轴角速度
		strLabelCaption[6] = "Acc   Z: "	+ String(MPU6050.Data.AccZ);										//Z轴加速度
	}
	else if (DisplayType == 3)
	{
		strLabelCaption[0] = "Angle X: "	+ String(MPU6050.Data.AngleX);										//X轴角度
		strLabelCaption[1] = "Gyro  X: "	+ String(MPU6050.Data.GyroX);										//X轴角速度
		strLabelCaption[2] = "Acc   X: "	+ String(MPU6050.Data.AccX);										//X轴加速度
		strLabelCaption[3] = "PWM: "		+ String(PWM);														//PWM
		strLabelCaption[4] = "PWM Turn: "	+ String(PWMTurn);													//转向PWM
		strLabelCaption[5] = "PWM Left: "	+ String(L_PWM);													//左轮PWM
		strLabelCaption[6] = "PWM Right: "	+ String(R_PWM);													//右轮PWM
	}
	else if (DisplayType == 4)
	{
		strLabelCaption[0] = "Power: "		+ String(Power_Percent) + "%";										//供电电量百分比
		strLabelCaption[1] = "Emergency: "	+ String(EmergencyStopStatus);										//急停状态
		strLabelCaption[2] = "AutoPatrol: "	+ String(AutoPatrol);												//是否自动巡逻
		strLabelCaption[3] = "Feedback: "	+ String(Feedback);													//是否反馈数据
		strLabelCaption[4] = "Feedback T: "	+ String(Feedback_Interval) + "ms";									//反馈数据间隔时间
		strLabelCaption[5] = "SerialBaud: "	+ String(WIRELESS_SERIAL_Baud);										//WIRELESS_SERIAL波特率
		strLabelCaption[6] = "BatteryNum: "	+ String(Power_BatteryNum);											//供电电池数量
	}
	else if (DisplayType == 5)
	{
		strLabelCaption[0] = "Power: "		+ String(Power_Percent) + "%";										//供电电量百分比
		strLabelCaption[1] = "kap: "		+ String(kap);														//角度环kp
		strLabelCaption[2] = "kai: "		+ String(kai);														//角度环ki
		strLabelCaption[3] = "kad: "		+ String(kad);														//角度环kd
		strLabelCaption[4] = "kaf: "		+ String(kaf);														//角度环kf
		strLabelCaption[5] = "kam: "		+ String(kam);														//角度环km
		strLabelCaption[6] = "kan: "		+ String(kan);														//角度环kn
	}
	else if (DisplayType == 6)
	{
		strLabelCaption[0] = "Power: "		+ String(Power_Percent) + "%";										//供电电量百分比
		strLabelCaption[1] = "ksp: "		+ String(ksp);														//速度环kp
		strLabelCaption[2] = "ksi: "		+ String(ksi);														//速度环ki
		strLabelCaption[3] = "ksd: "		+ String(ksd);														//速度环kd
		strLabelCaption[4] = "ksf: "		+ String(ksf);														//速度环kf
		strLabelCaption[5] = "ksm: "		+ String(ksm);														//速度环km
		strLabelCaption[6] = "ksn: "		+ String(ksn);														//速度环kn
	}
	else if (DisplayType == 7)
	{
		strLabelCaption[0] = "Power: "		+ String(Power_Percent) + "%";										//供电电量百分比
		strLabelCaption[1] = "ktp: "		+ String(ktp);														//转向环kp
		strLabelCaption[2] = "kti: "		+ String(kti);														//转向环ki
		strLabelCaption[3] = "ktd: "		+ String(ktd);														//转向环kd
		strLabelCaption[4] = "ktf: "		+ String(ktf);														//转向环kf
		strLabelCaption[5] = "ktm: "		+ String(ktm);														//转向环km
		strLabelCaption[6] = "ktn: "		+ String(ktn);														//转向环kn
	}
	else
	{
		DisplayType = 0;
	}

	if (DisplayType == 0)
	{
		if (boolDisplay)
		{
			boolDisplay = false;
			u8g2.clear();
		}
	}
	else
	{
		boolDisplay = true;
		u8g2.clearBuffer();
		for (int nIndex = 0; nIndex < 10; nIndex++) u8g2.drawUTF8(0, nIndex * 9, strLabelCaption[nIndex].c_str());
		u8g2.sendBuffer();
	}
}
//---------------------------------------------------------------------------
//获取电量百分比
int GetPowerPercent(const double fV)
{
	int nV = (int)trunc(fV * 1000);

	if      (nV >= 4200) return 100;
	else if (nV >= 4060) return (int)map(nV, 4060, 4200, 90, 100);
	else if (nV >= 3980) return (int)map(nV, 3980, 4060, 80,  90);
	else if (nV >= 3920) return (int)map(nV, 3920, 3980, 70,  80);
	else if (nV >= 3870) return (int)map(nV, 3870, 3920, 60,  70);
	else if (nV >= 3820) return (int)map(nV, 3820, 3870, 50,  60);
	else if (nV >= 3790) return (int)map(nV, 3790, 3820, 40,  50);
	else if (nV >= 3770) return (int)map(nV, 3770, 3790, 30,  40);
	else if (nV >= 3740) return (int)map(nV, 3740, 3770, 20,  30);
	else if (nV >= 3700) return (int)map(nV, 3700, 3740, 15,  20);
	else if (nV >= 3680) return (int)map(nV, 3680, 3700, 10,  15);
	else if (nV >= 3450) return (int)map(nV, 3450, 3680,  5,  10);
	else if (nV >= 3000) return (int)map(nV, 3000, 3450,  0,   5);
	else                 return 0;
}
//---------------------------------------------------------------------------
//获取供电电量百分比
bool GetPowerPercent(const int nLowPercent = 15)
{
	Power_Percent = GetPowerPercent((double)analogRead(Power_Pin) * 3.3 * 5 / Power_BatteryNum / 4095);
	return (Power_Percent >= nLowPercent);
}
//---------------------------------------------------------------------------
//获取姿态
bool GetAngle()
{
	if (! MPU6050.GetData()) return false;

	if (MPU6050.Data.AngleX < 0) MPU6050.Data.AngleX += 180; else if (MPU6050.Data.AngleX > 0) MPU6050.Data.AngleX -= 180;
	MPU6050.Data.GyroY = -MPU6050.Data.GyroY;
	MPU6050.Data.GyroZ = -MPU6050.Data.GyroZ;

	return true;
}
//---------------------------------------------------------------------------
//超声波测距（毫米）
int Ultrasonic_Ranging()
{
	//给US_Trig_Pin一个不小于10us的高电平脉冲信号，触发模块开始测距
	digitalWrite(US_Trig_Pin, LOW);
	delayMicroseconds(2);
	digitalWrite(US_Trig_Pin, HIGH);
	delayMicroseconds(10);
	digitalWrite(US_Trig_Pin, LOW);

	//模块测距完成后，US_Echo_Pin会返回一个高电平脉冲信号，脉冲信号持续的时间，就是从模块发出超声波开始，一直到接收到反弹回来的超声波为止的时间间隔（单位：微秒）
	unsigned long nPulseDuration = pulseIn(US_Echo_Pin, HIGH);

	//超声波也就是声音，在干燥、摄氏20度的空气中的传播速度大约为343米/秒，即0.343毫米/微秒，那么超声波来回走过的距离就是检测到的时间间隔*0.343（单位：毫米）
	double fDistance = (double)nPulseDuration * 0.343;
	fDistance /= 2;

	//因为探头是固定在小车上的，所以探测到的距离实际上是一个直角三角形的斜边距离，需要换算成邻边距离
	fDistance *= cos(abs(MPU6050.Data.AngleY) * DEG_TO_RAD);

	//返回测距结果
	SR04 = (int)trunc(fDistance);
	return SR04;
}
//---------------------------------------------------------------------------
//初始化超声波测距模块
bool Ultrasonic_Init()
{
	return (Ultrasonic_Ranging() > 0);
}
//---------------------------------------------------------------------------
//发送控制反馈
void Ctrl_Send(const String strCtrlType, const float fCtrlValue, const uint32_t nDelayTime = 1)
{
	String strSend = strCtrlType + "/" + String(fCtrlValue);

	uint8_t nChecksum = 0;
	for (int nIndex = 0; nIndex < strSend.length(); nIndex++) nChecksum += (uint8_t)strSend[nIndex];
	strSend += (char)nChecksum;

	if (USE_Communicate)
	{
		#ifdef USE_WIFI
			WIFI_Send(strSend);
		#endif

		#ifdef USE_BLE
			BLE_Send(strSend);
		#endif

		#ifdef USE_BT_SERIAL
			BT_SERIAL_Send(strSend);
		#endif

		#ifdef USE_WIRELESS_SERIAL
			WIRELESS_SERIAL_Send(strSend);
		#endif
	}

	if (nDelayTime > 0) vTaskDelay(nDelayTime);
}
//---------------------------------------------------------------------------
//反馈数据
void FeedbackData()
{
	if (Feedback == 0) return;

	Ctrl_Send("00", PIDTimeLoop);			//实际PID计算周期
	Ctrl_Send("01", MPU6050.Data.AngleX);	//X轴角度
	Ctrl_Send("02", MPU6050.Data.AngleY);	//Y轴角度
	Ctrl_Send("03", MPU6050.Data.AngleZ);	//Z轴角度
	Ctrl_Send("04", MPU6050.Data.GyroX);	//X轴角速度
	Ctrl_Send("05", MPU6050.Data.GyroY);	//Y轴角速度
	Ctrl_Send("06", MPU6050.Data.GyroZ);	//Z轴角速度
	Ctrl_Send("07", MPU6050.Data.AccX);		//X轴加速度
	Ctrl_Send("08", MPU6050.Data.AccY);		//Y轴加速度
	Ctrl_Send("09", MPU6050.Data.AccZ);		//Z轴加速度
	Ctrl_Send("10", SR04);					//超声波测距
	Ctrl_Send("11", PWM);					//PWM
	Ctrl_Send("12", PWMTurn);				//转向PWM
	Ctrl_Send("13", L_PWM);					//左轮PWM
	Ctrl_Send("14", R_PWM);					//右轮PWM
	Ctrl_Send("15", L_ENC_Count);			//左轮编码器计数
	Ctrl_Send("16", R_ENC_Count);			//右轮编码器计数
	Ctrl_Send("17", Power_Percent);			//供电电量百分比
	Ctrl_Send("21", Balance);				//平衡角度
	Ctrl_Send("22", GetMoveSpeed(true));	//移动速度
	Ctrl_Send("23", GetTurnSpeed(true));	//转向角速度
	Ctrl_Send("28", AutoPatrol);			//是否自动巡逻
	Ctrl_Send("29", Feedback);				//是否反馈数据
	Ctrl_Send("31", kap);					//角度环kp
	Ctrl_Send("32", kai);					//角度环ki
	Ctrl_Send("33", kad);					//角度环kd
	Ctrl_Send("34", kaf);					//角度环kf
	Ctrl_Send("35", kam);					//角度环km
	Ctrl_Send("36", kan);					//角度环kn
	Ctrl_Send("41", ksp);					//速度环kp
	Ctrl_Send("42", ksi);					//速度环ki
	Ctrl_Send("43", ksd);					//速度环kd
	Ctrl_Send("44", ksf);					//速度环kf
	Ctrl_Send("45", ksm);					//速度环km
	Ctrl_Send("46", ksn);					//速度环kn
	Ctrl_Send("51", ktp);					//转向环kp
	Ctrl_Send("52", kti);					//转向环ki
	Ctrl_Send("53", ktd);					//转向环kd
	Ctrl_Send("54", ktf);					//转向环kf
	Ctrl_Send("55", ktm);					//转向环km
	Ctrl_Send("56", ktn);					//转向环kn
	Ctrl_Send("81", EmergencyStopStatus);	//急停状态
	Ctrl_Send("91", WIRELESS_SERIAL_Baud);	//WIRELESS_SERIAL波特率
	Ctrl_Send("92", Feedback_Interval);		//反馈数据间隔时间
	Ctrl_Send("93", Power_BatteryNum);		//供电电池数量
}
//---------------------------------------------------------------------------
//处理控制内容
void Ctrl_Receive_Process(const String strCtrlType, const float fCtrlValue)
{
	if      (strCtrlType == "81") {EmergencyStopStatus	= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//急停状态
	else if (strCtrlType == "A0") {MoveSpeed			= 0;          TurnSpeed = 0;                      }		//停止
	else if (strCtrlType == "A1") {Balance				= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//平衡角度
	else if (strCtrlType == "A2") {MoveSpeed			= fCtrlValue;                                     }		//移动速度
	else if (strCtrlType == "A3") {TurnSpeed			= fCtrlValue;                                     }		//转向角速度
	else if (strCtrlType == "A8") {AutoPatrol			= fCtrlValue; Ctrl_Receive_Process("A0", 0);      }		//是否自动巡逻
	else if (strCtrlType == "A9") {Feedback				= fCtrlValue;                                     }		//是否反馈数据
	else if (strCtrlType == "B1") {kap					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//角度环kp
	else if (strCtrlType == "B2") {kai					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//角度环ki
	else if (strCtrlType == "B3") {kad					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//角度环kd
	else if (strCtrlType == "B4") {kaf					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//角度环kf
	else if (strCtrlType == "B5") {kam					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//角度环km
	else if (strCtrlType == "B6") {kan					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//角度环kn
	else if (strCtrlType == "C1") {ksp					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//速度环kp
	else if (strCtrlType == "C2") {ksi					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//速度环ki
	else if (strCtrlType == "C3") {ksd					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//速度环kd
	else if (strCtrlType == "C4") {ksf					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//速度环kf
	else if (strCtrlType == "C5") {ksm					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//速度环km
	else if (strCtrlType == "C6") {ksn					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//速度环kn
	else if (strCtrlType == "D1") {ktp					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//转向环kp
	else if (strCtrlType == "D2") {kti					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//转向环ki
	else if (strCtrlType == "D3") {ktd					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//转向环kd
	else if (strCtrlType == "D4") {ktf					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//转向环kf
	else if (strCtrlType == "D5") {ktm					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//转向环km
	else if (strCtrlType == "D6") {ktn					= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//转向环kn
	else if (strCtrlType == "F1") {WIRELESS_SERIAL_Baud	= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//WIRELESS_SERIAL波特率
	else if (strCtrlType == "F2") {Feedback_Interval	= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//反馈数据间隔时间
	else if (strCtrlType == "F3") {Power_BatteryNum		= fCtrlValue; WriteParam(strCtrlType, fCtrlValue);}		//供电电池数量
}
//---------------------------------------------------------------------------
//接收控制内容
void Ctrl_Receive()
{
	while (USE_Communicate)
	{
		String strReceive = "";

		bool boolSerial = Serial.available();
		if (boolSerial) strReceive = Serial.readStringUntil('\n');

		#ifdef USE_WIFI
			if (strReceive.isEmpty()) strReceive = WIFI_Receive();
		#endif

		#ifdef USE_BLE
			if (strReceive.isEmpty()) strReceive = BLE_Receive();
		#endif

		#ifdef USE_BT_SERIAL
			if (strReceive.isEmpty()) strReceive = BT_SERIAL_Receive();
		#endif

		#ifdef USE_WIRELESS_SERIAL
			if (strReceive.isEmpty()) strReceive = WIRELESS_SERIAL_Receive();
		#endif

		if (strReceive.isEmpty())
		{
			break;
		}
		else
		{
			if (strReceive.substring(2, 3) == "/")
			{
				if (boolSerial) strReceive += " ";
				uint8_t nChecksum = 0;
				for (int nIndex = 0; nIndex < strReceive.length() - 1; nIndex++) nChecksum += (uint8_t)strReceive[nIndex];

				if (boolSerial || (nChecksum == (uint8_t)strReceive[strReceive.length() - 1]))
				{
					String strCtrlType = strReceive.substring(0, 2);
					float fCtrlValue = strReceive.substring(3, strReceive.length() - 1).toFloat();
					Ctrl_Receive_Process(strCtrlType, fCtrlValue);
//					Ctrl_Send(strCtrlType, fCtrlValue, 0);
				}
			}
		}
	}
}
//---------------------------------------------------------------------------
//自动巡逻
void doAutoPatrol()
{
	//自主巡逻
	if (AutoPatrol == 1)
	{
		//模式标志
		static bool boolMove = false;  //巡逻模式
		static bool boolAvoi = false;  //避障模式

		//模式起始时间
		static unsigned long nMoveTime = millis();
		static unsigned long nAvoiTime = millis();

		//巡逻模式判断
		if (boolMove)
		{
			//如果巡逻模式持续60-90秒以上，并且不处于避障模式，则切换为暂停模式
			if ((millis() - nMoveTime > random(60 * 1000, 90 * 1000)) && (! boolAvoi))
			{
				boolMove = false;
				nMoveTime = millis();
			}
		}
		else
		{
			//如果暂停模式持续10-20秒以上，则切换为巡逻模式
			if (millis() - nMoveTime > random(10 * 1000, 20 * 1000))
			{
				boolMove = true;
				nMoveTime = millis();
			}
		}

		//巡逻模式
		if (boolMove)
		{
			//避障模式
			if (boolAvoi)
			{
				//如果避障模式持续3-10秒以上，并且障碍距离大于1000mm，则关闭避障模式
				if ((millis() - nAvoiTime > random(3 * 1000, 10 * 1000)) && (SR04 > 1000)) boolAvoi = false;
			}
			else
			{
				//正常巡逻（障碍距离大于1000mm）
				if (SR04 > 1000)
				{
					MoveSpeed = MoveSpeedMax / 4;
					TurnSpeed = 0;
				}
				//减速巡逻（障碍距离大于400mm）
				else if (SR04 > 400)
				{
					MoveSpeed = map(SR04, 400, 1000, MoveSpeedMax / 40, MoveSpeedMax / 4);
					TurnSpeed = 0;
				}
				//打开避障模式（障碍距离小于400mm）
				else
				{
					MoveSpeed = 0;
					TurnSpeed = ((random(0, 2) == 0)?(-TurnSpeedMax / 6):(TurnSpeedMax / 6));

					boolAvoi = true;
					nAvoiTime = millis();
				}
			}
		}
		//暂停模式
		else
		{
			MoveSpeed = 0;
			TurnSpeed = 0;
		}
	}
	//跟随巡逻
	else if (AutoPatrol == 2)
	{
		//暂停（障碍距离大于1500mm）
		if (SR04 > 1500)
		{
			MoveSpeed = 0;
			TurnSpeed = 0;
		}
		//正常跟随（障碍距离大于1000mm）
		else if (SR04 > 1000)
		{
			MoveSpeed = MoveSpeedMax / 4;
			TurnSpeed = 0;
		}
		//减速跟随（障碍距离大于500mm）
		else if (SR04 > 500)
		{
			MoveSpeed = map(SR04, 500, 1000, MoveSpeedMax / 40, MoveSpeedMax / 4);
			TurnSpeed = 0;
		}
		//停留（障碍距离大于300mm）
		else if (SR04 > 300)
		{
			MoveSpeed = 0;
			TurnSpeed = 0;
		}
		//避障（障碍距离小于300mm）
		else
		{
			MoveSpeed = -MoveSpeedMax / 4;
			TurnSpeed = 0;
		}
	}
	else
	{
		AutoPatrol = 0;
	}
}
//---------------------------------------------------------------------------
//检测按钮
uint8_t ButtonClick()
{
	static bool boolBtnDown = false;
	static uint8_t nBtnClickNum = 0;

	static unsigned long nLastTime_BtnDown  = millis() - 1000;
	static unsigned long nLastTime_BtnClick = millis() - 1000;

	unsigned long now = millis();
	uint8_t nResult = 0;

	//判断按钮是否按下或抬起
	if (digitalRead(EmeStop_Pin) == (boolBtnDown?HIGH:LOW))
	{
		vTaskDelay(20);  //去抖
		if (digitalRead(EmeStop_Pin) == (boolBtnDown?HIGH:LOW))
		{
			boolBtnDown = (! boolBtnDown);

			if (boolBtnDown)
			{
				nLastTime_BtnDown = now;
			}
			else
			{
				//如果按钮从按下到抬起的时间小于300ms，则认为是点击
				if (now - nLastTime_BtnDown < 300)
				{
					//如果两次点击间隔时间小于300ms，则认为是多次点击
					if (now - nLastTime_BtnClick < 300) nBtnClickNum++; else nBtnClickNum = 1;
					nLastTime_BtnClick = now;
				}
				else
				{
					nBtnClickNum = 0;
				}
			}
		}
	}

	//如果已经是3次点击，或者上次点击时间是300ms之前，则返回点击次数
	if ((nBtnClickNum >= 3) || (now - nLastTime_BtnClick >= 300))
	{
		nResult = nBtnClickNum;
		nBtnClickNum = 0;
	}

	return nResult;
}
//---------------------------------------------------------------------------
//编码器中断（IRAM_ATTR修饰词：函数常驻RAM中，加快执行速度）
void IRAM_ATTR L_ENA_Pin_Interrupt() {portENTER_CRITICAL_ISR(&L_ENC_MUX); if (digitalRead(L_ENA_Pin) == digitalRead(L_ENB_Pin)) L_ENC++; else L_ENC--; portEXIT_CRITICAL_ISR(&L_ENC_MUX);}
void IRAM_ATTR L_ENB_Pin_Interrupt() {portENTER_CRITICAL_ISR(&L_ENC_MUX); if (digitalRead(L_ENA_Pin) != digitalRead(L_ENB_Pin)) L_ENC++; else L_ENC--; portEXIT_CRITICAL_ISR(&L_ENC_MUX);}
void IRAM_ATTR R_ENA_Pin_Interrupt() {portENTER_CRITICAL_ISR(&R_ENC_MUX); if (digitalRead(R_ENA_Pin) == digitalRead(R_ENB_Pin)) R_ENC--; else R_ENC++; portEXIT_CRITICAL_ISR(&R_ENC_MUX);}
void IRAM_ATTR R_ENB_Pin_Interrupt() {portENTER_CRITICAL_ISR(&R_ENC_MUX); if (digitalRead(R_ENA_Pin) != digitalRead(R_ENB_Pin)) R_ENC--; else R_ENC++; portEXIT_CRITICAL_ISR(&R_ENC_MUX);}
//---------------------------------------------------------------------------
//急停判断
bool EmergencyStop()
{
	static bool boolEmergencyStop_Last = true;
	bool boolEmergencyStop = false;
	unsigned long now = millis();

	//左电机卡死
	if (! boolEmergencyStop)
	{
		static unsigned long nLastTime = millis();
		bool boolStop = ((abs(L_PWM) > PWMMax / 5) && (abs(L_ENC_Cycle) < map(abs(L_PWM), 0, PWMMax, 0, ENCMax / 2)));
		if (boolStop) boolEmergencyStop = (now - nLastTime > 300); else nLastTime = now;
		if (boolEmergencyStop && (EmergencyStopStatus == 0)) Ctrl_Receive_Process("81", 1);
	}

	//右电机卡死
	if (! boolEmergencyStop)
	{
		static unsigned long nLastTime = millis();
		bool boolStop = ((abs(R_PWM) > PWMMax / 5) && (abs(R_ENC_Cycle) < map(abs(R_PWM), 0, PWMMax, 0, ENCMax / 2)));
		if (boolStop) boolEmergencyStop = (now - nLastTime > 300); else nLastTime = now;
		if (boolEmergencyStop && (EmergencyStopStatus == 0)) Ctrl_Receive_Process("81", 2);
	}

	//左电机空转
	if (! boolEmergencyStop)
	{
		static unsigned long nLastTime = millis();
		bool boolStop = ((abs(L_PWM) == PWMMax) && (GetMoveSpeed(true) == 0) && (GetTurnSpeed(true) == 0));
		if (boolStop) boolEmergencyStop = (now - nLastTime > 300); else nLastTime = now;
		if (boolEmergencyStop && (EmergencyStopStatus == 0)) Ctrl_Receive_Process("81", 3);
	}

	//右电机空转
	if (! boolEmergencyStop)
	{
		static unsigned long nLastTime = millis();
		bool boolStop = ((abs(R_PWM) == PWMMax) && (GetMoveSpeed(true) == 0) && (GetTurnSpeed(true) == 0));
		if (boolStop) boolEmergencyStop = (now - nLastTime > 300); else nLastTime = now;
		if (boolEmergencyStop && (EmergencyStopStatus == 0)) Ctrl_Receive_Process("81", 4);
	}

	//急停按钮按下
	if (! boolEmergencyStop)
	{
		static unsigned long nLastTime = millis();
		bool boolStop = (digitalRead(EmeStop_Pin) == LOW);
		if (boolStop) boolEmergencyStop = (now - nLastTime >= 0); else nLastTime = now;
		if (boolEmergencyStop && (EmergencyStopStatus == 0)) Ctrl_Receive_Process("81", 5);
	}

	//Y角度偏差过大
	if (! boolEmergencyStop)
	{
		static unsigned long nLastTime = millis();
		bool boolStop = (abs(MPU6050.Data.AngleY) > 60);
		if (boolStop) boolEmergencyStop = (now - nLastTime >= 0); else nLastTime = now;
		if (boolEmergencyStop && (EmergencyStopStatus == 0)) Ctrl_Receive_Process("81", 6);
	}

	//X角度偏差过大
	if (! boolEmergencyStop)
	{
		static unsigned long nLastTime = millis();
		bool boolStop = (abs(MPU6050.Data.AngleX) > 60);
		if (boolStop) boolEmergencyStop = (now - nLastTime > 100); else nLastTime = now;
		if (boolEmergencyStop && (EmergencyStopStatus == 0)) Ctrl_Receive_Process("81", 7);
	}

	//判断是否解除急停状态
	if (boolEmergencyStop_Last && (! boolEmergencyStop))
	{
		static unsigned long nLastTime = millis();

		if (abs(MPU6050.Data.AngleY - Balance) < 5)
		{
			if (now - nLastTime > 100)
			{
				boolEmergencyStop_Last = boolEmergencyStop;
				Ctrl_Receive_Process("81", 0);
			}
			else
			{
				boolEmergencyStop = boolEmergencyStop_Last;
			}
		}
		else
		{
			boolEmergencyStop = boolEmergencyStop_Last;
			nLastTime = now;
		}
	}
	else
	{
		boolEmergencyStop_Last = boolEmergencyStop;
	}

	return boolEmergencyStop;
}
//---------------------------------------------------------------------------
//角度环计算
float PID_Calc_Angle(const bool boolEmergencyStop, const float fOther1 = 0, const float fOther2 = 0, const float fOther3 = 0)
{
	static float fError = 0;
	static float fIntegral = 0;

	//急停判断
	if (boolEmergencyStop)
	{
		fError = 0;
		fIntegral = 0;
		return 0;
	}

	//其他角度介入
	float fOther = fOther1 + fOther2 + fOther3;
	fOther = constrain(fOther, -15, 15);

	//误差（低通滤波，过滤高频噪声）
	fError *= (1 - kaf);
	fError += ((MPU6050.Data.AngleY - Balance - fOther) * kaf);

	//积分
	fIntegral += fError;

	//微分
	float fDifferential = MPU6050.Data.GyroY - 0;  //加速前进、后退，可以在这里修改偏差

	//PID计算
	float fPID = kap * fError + kai * fIntegral + kad * fDifferential;
	if ((GetMoveSpeed() == 0) && (GetTurnSpeed() == 0)) fPID *= kam; else fPID *= kan;
	return -fPID;
}
//---------------------------------------------------------------------------
//速度环计算
float PID_Calc_Speed(const bool boolEmergencyStop)
{
	static float fError = 0;
	static float fIntegral = 0;
	static float fError_Last = fError;

	//急停判断
	if (boolEmergencyStop)
	{
		fError = 0;
		fIntegral = 0;
		fError_Last = 0;
		return 0;
	}

	//误差（低通滤波，过滤高频噪声）
	fError *= (1 - ksf);
	fError += ((L_ENC_Cycle + R_ENC_Cycle) * ksf);

	//积分
	fIntegral += fError;
	fIntegral -= GetMoveSpeed();

	//微分
	float fDifferential = fError - fError_Last;
	fError_Last = fError;

	//PID计算
	float fPID = ksp * fError + ksi * fIntegral + ksd * fDifferential;
	if ((GetMoveSpeed() == 0) && (GetTurnSpeed() == 0)) fPID *= ksm; else fPID *= ksn;
	fPID /= 10.0;  //PID系数是放大10倍的
	return fPID;
}
//---------------------------------------------------------------------------
//转向环计算
float PID_Calc_Turn(const bool boolEmergencyStop)
{
	static float fError = 0;
	static float fIntegral = 0;
	static float fError_Last = fError;

	//急停判断
	if (boolEmergencyStop)
	{
		fError = 0;
		fIntegral = 0;
		fError_Last = 0;
		return 0;
	}

	//误差（低通滤波，过滤高频噪声）
	fError *= (1 - ktf);
	fError += (MPU6050.Data.GyroZ * ktf);

	//积分
	fIntegral += fError;
	fIntegral -= GetTurnSpeed();

	//微分
	float fDifferential = fError - fError_Last;
	fError_Last = fError;

	//PID计算
	float fPID = ktp * fError + kti * fIntegral + ktd * fDifferential;
	if ((GetMoveSpeed() == 0) && (GetTurnSpeed() == 0)) fPID *= ktm; else fPID *= ktn;
	fPID /= 10.0;  //PID系数是放大10倍的
	return fPID;
}
//---------------------------------------------------------------------------
//PWM计算
void PWM_Calc()
{
	//实际PID计算周期
	static unsigned long nLastTime = millis();
	unsigned long now = millis();
	PIDTimeLoop = (int)(now - nLastTime);
	nLastTime = now;

	//获取编码器计数
	portENTER_CRITICAL_ISR(&L_ENC_MUX); L_ENC_Count += L_ENC; L_ENC_Cycle = L_ENC; L_ENC = 0; portEXIT_CRITICAL_ISR(&L_ENC_MUX);
	portENTER_CRITICAL_ISR(&R_ENC_MUX); R_ENC_Count += R_ENC; R_ENC_Cycle = R_ENC; R_ENC = 0; portEXIT_CRITICAL_ISR(&R_ENC_MUX);

	//获取急停判断
	bool boolEmergencyStop = EmergencyStop();

	//PID计算PWM
	PWM = (int)(PID_Calc_Angle(boolEmergencyStop) + PID_Calc_Speed(boolEmergencyStop));
	PWMTurn = (int)PID_Calc_Turn(boolEmergencyStop);

	//急停判断
	if (boolEmergencyStop)
	{
		L_PWM = 0;	L_ENC_Count = 0;
		R_PWM = 0;	R_ENC_Count = 0;
	}
	else
	{
		L_PWM = PWM;
		R_PWM = PWM;

		//如果没有移动速度，则原地转向
		if (GetMoveSpeed() == 0)
		{
			L_PWM += (PWMTurn / 2);
			R_PWM -= (PWMTurn / 2);
		}
		else
		{
			if (PWMTurn > 0) L_PWM += PWMTurn;
			if (PWMTurn < 0) R_PWM -= PWMTurn;
		}

		//限制PWM最大值
		L_PWM = constrain(L_PWM, -PWMMax, PWMMax);
		R_PWM = constrain(R_PWM, -PWMMax, PWMMax);
	}
}
//---------------------------------------------------------------------------
//PWM输出
void PWM_Ctrl(const int nL, const int nR, const bool boolNoStop = false)
{
	bool boolCond1, boolCond2;

	if (boolNoStop)
	{
		boolCond1 = (nL >= 0);
		boolCond2 = (nL <= 0);
	}
	else
	{
		boolCond1 = (nL > 0);
		boolCond2 = (nL < 0);
	}

	digitalWrite(L_IN1_Pin, boolCond1?LOW:HIGH);
	digitalWrite(L_IN2_Pin, boolCond2?LOW:HIGH);
	analogWrite(L_PWM_Pin, abs(nL));

	if (boolNoStop)
	{
		boolCond1 = (nR <= 0);
		boolCond2 = (nR >= 0);
	}
	else
	{
		boolCond1 = (nR < 0);
		boolCond2 = (nR > 0);
	}

	digitalWrite(R_IN1_Pin, boolCond1?LOW:HIGH);
	digitalWrite(R_IN2_Pin, boolCond2?LOW:HIGH);
	analogWrite(R_PWM_Pin, abs(nR));
}
//---------------------------------------------------------------------------
