//---------------------------------------------------------------------------
#include <BLEDevice.h>
//---------------------------------------------------------------------------
BLEUUID UUID_SERVICE("a8df5590-9515-11ea-bb37-0242ac130002");
BLEUUID UUID_RW     ("a8df5798-9515-11ea-bb37-0242ac130002");
//---------------------------------------------------------------------------
BLEAdvertisedDevice*		pBLEDevice;
BLERemoteCharacteristic*	pBLEReadWrite;
//---------------------------------------------------------------------------
bool BLE_doConnect = false;  //是否可以连接
bool Connected     = false;  //是否连接成功
//---------------------------------------------------------------------------
class BLE_Scan_Callback: public BLEAdvertisedDeviceCallbacks
{
	void onResult(BLEAdvertisedDevice advertisedDevice)
	{
		Serial.print("Scaned: ");
		Serial.println(advertisedDevice.toString().c_str());

		//判断UUID_SERVICE是否符合
		if (advertisedDevice.haveServiceUUID() && advertisedDevice.isAdvertisingService(UUID_SERVICE))
		{
			Serial.println("UUID_SERVICE ok");

			//停止扫描
			BLEDevice::getScan()->stop();

			//建立BLE设备
			pBLEDevice = new BLEAdvertisedDevice(advertisedDevice);

			//初始化标志位
			BLE_doConnect = true;
		}
	}
};
//---------------------------------------------------------------------------
class BLE_Client_Callback: public BLEClientCallbacks
{
	void onConnect(BLEClient* pBLEClient) {}
	void onDisconnect(BLEClient* pBLEClient)
	{
		Connected = false;
		Serial.println("Disconnect");
	}
};
//---------------------------------------------------------------------------
String BLEReceive = "";
void BLE_Notify_Callback(BLERemoteCharacteristic* pBLERemoteCharacteristic, uint8_t* pData, size_t length, bool isNotify)
{
	std::string strValue = pBLERemoteCharacteristic->readValue();
	for (int nIndex = 0; nIndex < strValue.length(); nIndex++) BLEReceive += String(strValue[nIndex]);
}
//---------------------------------------------------------------------------
bool Connect()
{
	Serial.print("Connect address: ");
	Serial.println(pBLEDevice->getAddress().toString().c_str());

	//建立BLE客户端
	BLEClient* pBLEClient = BLEDevice::createClient();
	pBLEClient->setClientCallbacks(new BLE_Client_Callback());
	if (pBLEClient->connect(pBLEDevice))
	{
		Serial.println("Connect ok");
	}
	else
	{
		Serial.println("Connect failure");
		return false;
	}

	//判断UUID_SERVICE是否符合
	BLERemoteService* pBLEService = pBLEClient->getService(UUID_SERVICE);
	if (pBLEService != nullptr)
	{
		Serial.println("UUID_SERVICE ok");
	}
	else
	{
		Serial.println("UUID_SERVICE failure");
		pBLEClient->disconnect();
		return false;
	}

	//判断UUID_RW是否符合
	pBLEReadWrite = pBLEService->getCharacteristic(UUID_RW);
	if (pBLEReadWrite != nullptr)
	{
		Serial.println("UUID_RW ok");
	}
	else
	{
		Serial.println("UUID_RW failure");
		pBLEClient->disconnect();
		return false;
	}

	//初始化Notify回调函数
//	if (pBLEReadWrite->canNotify()) pBLEReadWrite->registerForNotify(BLE_Notify_Callback);

	//初始化标志位
	Connected = true;
	Serial.println("Connected");
	return true;
}
//---------------------------------------------------------------------------
bool BLE_KeepConnect()
{
	if (! Connected) BLE_Init();
	return Connected;
}
//---------------------------------------------------------------------------
bool BLE_Init()
{
	//初始化BLE设备
	BLEDevice::init("");

    //建立BLE扫描
    BLEScan* pBLEScan = BLEDevice::getScan();
    pBLEScan->setAdvertisedDeviceCallbacks(new BLE_Scan_Callback());
    pBLEScan->setActiveScan(true);	//设置为主动扫描
    pBLEScan->setWindow(500);		//单个扫描时间：500ms
    pBLEScan->setInterval(1000);	//扫描间隔时间：1000ms
	pBLEScan->start(300, false);	//总共扫描时间：300s

	//连接Server
	unsigned long nTime = millis();
	while (millis() - nTime < 300 * 1000) if (BLE_doConnect) return Connect();
	return false;
}
//---------------------------------------------------------------------------
void BLE_Write(const String strValue)
{
	std::string strSend = "";
	strSend.append(strValue.c_str());
	pBLEReadWrite->writeValue(strSend);
}
//---------------------------------------------------------------------------
String BLE_Receive()
{
	String strReceive = "";

	if (! Connected) return strReceive;
	if (! pBLEReadWrite->canRead()) return strReceive;
//	if (! pBLEReadWrite->canNotify()) return strReceive;

	static std::string strValue_Last = "";
	std::string strValue = pBLEReadWrite->readValue();
	if (strValue != strValue_Last)
	{
		for (int nIndex = 0; nIndex < strValue.length(); nIndex++) BLEReceive += String(strValue[nIndex]);
		strValue_Last = strValue;
	}

	int nIndex = BLEReceive.indexOf("\n");
	if (nIndex >= 0)
	{
		strReceive = BLEReceive.substring(0, nIndex);
		BLEReceive = BLEReceive.substring(nIndex + 1, BLEReceive.length());
	}

	return strReceive;
}
//---------------------------------------------------------------------------
void BLE_Send(String strSend)
{
	if (! Connected) return;
	if (! pBLEReadWrite->canWrite()) return;

	strSend += "\n";
	BLE_Write(strSend);
}
//---------------------------------------------------------------------------
