//---------------------------------------------------------------------------
#include "Define.h"
//---------------------------------------------------------------------------
//读取参数
float ReadParam(const String strCtrlType)
{
	char c[4];
	strCtrlType.toCharArray(c, sizeof(c));
	uint16_t nCtrlType;
	sscanf(c, "%x", &nCtrlType);

	float fResult = 0;
	if (nCtrlType < 0x100) fResult = EEPROM.readFloat(nCtrlType * sizeof(float) + 2100);  //0-2047被Blinker占用
	if (isnan(fResult)) fResult = 0;
	return fResult;
}
//---------------------------------------------------------------------------
//读取全部参数
void ReadAllParams()
{
	WIRELESS_SERIAL_Baud	= ReadParam("F1");	Serial.print("WIRELESS_SERIAL_Baud: ");	Serial.println(WIRELESS_SERIAL_Baud);	//WIRELESS_SERIAL波特率
	Power_BatteryNum		= 1;				Serial.print("Power_BatteryNum: ");		Serial.println(Power_BatteryNum);		//供电电池数量
}
//---------------------------------------------------------------------------
//保存参数
void WriteParam(const String strCtrlType, const float fCtrlValue)
{
	char c[4];
	strCtrlType.toCharArray(c, sizeof(c));
	uint16_t nCtrlType;
	sscanf(c, "%x", &nCtrlType);

	if (nCtrlType < 0x100)
	{
		EEPROM.writeFloat(nCtrlType * sizeof(float) + 2100, fCtrlValue);  //0-2047被Blinker占用
		EEPROM.commit();
	}
}
//---------------------------------------------------------------------------
//初始化显示屏
void Display_Init()
{
	//初始化显示屏背光引脚
	pinMode(TFT_BLK_Pin, OUTPUT);
	digitalWrite(TFT_BLK_Pin, HIGH);

	//初始化显示屏
	tft.init(135, 240);				//分辨率240x135
	tft.setRotation(0);				//0：0度；1：90度；2：180度；3：270度
	tft.setFont(NULL);				//使用内置6x8字体
	tft.setTextSize(1, 1);			//字体放大倍数
	tft.setTextWrap(false);			//不自动换行
	tft.fillScreen(ST77XX_BLACK);	//ST77XX_BLACK、ST77XX_WHITE、ST77XX_RED、ST77XX_GREEN、ST77XX_BLUE、ST77XX_CYAN、ST77XX_MAGENTA、ST77XX_YELLOW、ST77XX_ORANGE、0xFFFF
}
//---------------------------------------------------------------------------
//显示标签
void Display_Label(const int nLabel, const String strCaption, const uint16_t nColor = ST77XX_WHITE, const bool boolSerialPrint = true)
{
	tft.setCursor(0, nLabel * 9);
	tft.setTextColor(nColor, ST77XX_BLACK);

	if (strCaption.length() < 22)
	{
		tft.print(strCaption);
		for (int nIndex = 0; nIndex < 22 - strCaption.length(); nIndex++) tft.print(" ");
	}
	else
	{
		tft.print(strCaption.substring(0, 22));
	}

	if (boolSerialPrint)
	{
		Serial.print(nLabel);
		Serial.print(". ");
		Serial.println(strCaption);
	}
}
//---------------------------------------------------------------------------
//显示内容
void Display_Content(const int nLabel, String strCaption, const uint16_t nColor, const float fCtrlValue, const String strUnit, const bool boolInt)
{
	strCaption += ": ";
	if (boolInt) strCaption += String((int)fCtrlValue); else strCaption += String(fCtrlValue);
	strCaption += strUnit;
	Display_Label(nLabel, strCaption, nColor, false);
	vTaskDelay(1);
}
//---------------------------------------------------------------------------
//显示内容
void Display_Content()
{
	static bool boolDisplay = true;

	if (DisplayType == 1)
	{
		Display_Content( 0, "Joystick Power",	ST77XX_RED,		Power_Percent,			"%",	true);		//Joystick 供电电量百分比
		Display_Content( 1, "Joystick Y",		ST77XX_WHITE,	MPU6050.Data.AngleY,	"",		false);		//Joystick Y轴角度
		Display_Content( 2, "Joystick X",		ST77XX_WHITE,	MPU6050.Data.AngleX,	"",		false);		//Joystick X轴角度
		Display_Content( 3, "Power",			ST77XX_RED,		F_Power_Percent,		"%",	true);		//供电电量百分比
		Display_Content( 4, "PID Time",			ST77XX_WHITE,	F_PIDTimeLoop,			"ms",	true);		//实际PID计算周期（毫秒）
		Display_Content( 5, "Angle X",			ST77XX_WHITE,	F_AngleX,				"",		false);		//X轴角度
		Display_Content( 6, "Angle Y",			ST77XX_WHITE,	F_AngleY,				"",		false);		//Y轴角度
		Display_Content( 7, "Angle Z",			ST77XX_WHITE,	F_AngleZ,				"",		false);		//Z轴角度
		Display_Content( 8, "Gyro X",			ST77XX_GREEN,	F_GyroX,				"",		false);		//X轴角速度
		Display_Content( 9, "Gyro Y",			ST77XX_GREEN,	F_GyroY,				"",		false);		//Y轴角速度
		Display_Content(10, "Gyro Z",			ST77XX_GREEN,	F_GyroZ,				"",		false);		//Z轴角速度
		Display_Content(11, "Acc X",			ST77XX_WHITE,	F_AccX,					"",		false);		//X轴加速度
		Display_Content(12, "Acc Y",			ST77XX_WHITE,	F_AccY,					"",		false);		//Y轴加速度
		Display_Content(13, "Acc Z",			ST77XX_WHITE,	F_AccZ,					"",		false);		//Z轴加速度
		Display_Content(14, "SR04",				ST77XX_RED,		F_SR04,					"mm",	true);		//SR04超声波测距
		Display_Content(15, "PWM",				ST77XX_WHITE,	F_PWM,					"",		true);		//PWM
		Display_Content(16, "PWM Turn",			ST77XX_WHITE,	F_PWMTurn,				"",		true);		//转向PWM
		Display_Content(17, "PWM Left",			ST77XX_WHITE,	F_L_PWM,				"",		true);		//左轮PWM
		Display_Content(18, "PWM Right",		ST77XX_WHITE,	F_R_PWM,				"",		true);		//右轮PWM
		Display_Content(19, "ENC Left",			ST77XX_GREEN,	F_L_ENC_Count,			"",		true);		//左轮编码器计数总计
		Display_Content(20, "ENC Right",		ST77XX_GREEN,	F_R_ENC_Count,			"",		true);		//右轮编码器计数总计
		Display_Content(21, "Emergency",		ST77XX_WHITE,	F_EmergencyStopStatus,	"",		true);		//急停状态
		Display_Content(22, "Move Speed",		ST77XX_YELLOW,	F_MoveSpeed,			"",		false);		//移动速度（一个PID计算周期内左右编码器计数之和）
		Display_Content(23, "Turn Speed",		ST77XX_YELLOW,	F_TurnSpeed,			"",		false);		//转向角速度（°/s）
		Display_Content(24, "Auto Patrol",		ST77XX_WHITE,	F_AutoPatrol,			"",		true);		//是否自动巡逻
	}
	else if (DisplayType == 2)
	{
		Display_Content( 0, "Balance",			ST77XX_WHITE,	F_Balance,				"",		false);		//平衡角度（°）
		Display_Content( 1, "Serial Baud",		ST77XX_WHITE,	F_WIRELESS_SERIAL_Baud,	"",		true);		//WIRELESS_SERIAL波特率
		Display_Content( 2, "Feedback",			ST77XX_WHITE,	F_Feedback_Interval,	"ms",	true);		//反馈数据间隔时间
		Display_Content( 3, "Battery Num",		ST77XX_WHITE,	F_Power_BatteryNum,		"",		true);		//供电电池数量
		Display_Content( 4, "kap",				ST77XX_GREEN,	F_kap,					"",		false);		//角度环kap
		Display_Content( 5, "kai",				ST77XX_WHITE,	F_kai,					"",		false);		//角度环kai
		Display_Content( 6, "kad",				ST77XX_WHITE,	F_kad,					"",		false);		//角度环kad
		Display_Content( 7, "kaf",				ST77XX_WHITE,	F_kaf,					"",		false);		//角度环kaf
		Display_Content( 8, "kam",				ST77XX_WHITE,	F_kam,					"",		false);		//角度环kam
		Display_Content( 9, "kan",				ST77XX_WHITE,	F_kan,					"",		false);		//角度环kan
		Display_Content(10, "ksp",				ST77XX_GREEN,	F_ksp,					"",		false);		//速度环ksp
		Display_Content(11, "ksi",				ST77XX_WHITE,	F_ksi,					"",		false);		//速度环ksi
		Display_Content(12, "ksd",				ST77XX_WHITE,	F_ksd,					"",		false);		//速度环ksd
		Display_Content(13, "ksf",				ST77XX_WHITE,	F_ksf,					"",		false);		//速度环ksf
		Display_Content(14, "ksm",				ST77XX_WHITE,	F_ksm,					"",		false);		//速度环ksm
		Display_Content(15, "ksn",				ST77XX_WHITE,	F_ksn,					"",		false);		//速度环ksn
		Display_Content(16, "ktp",				ST77XX_GREEN,	F_ktp,					"",		false);		//转向环ktp
		Display_Content(17, "kti",				ST77XX_WHITE,	F_kti,					"",		false);		//转向环kti
		Display_Content(18, "ktd",				ST77XX_WHITE,	F_ktd,					"",		false);		//转向环ktd
		Display_Content(19, "ktf",				ST77XX_WHITE,	F_ktf,					"",		false);		//转向环ktf
		Display_Content(20, "ktm",				ST77XX_WHITE,	F_ktm,					"",		false);		//转向环ktm
		Display_Content(21, "ktn",				ST77XX_WHITE,	F_ktn,					"",		false);		//转向环ktn
		Display_Content(22, "Feedback",			ST77XX_WHITE,	F_Feedback,				"",		true);		//是否反馈数据
		Display_Content(23, "Received",			ST77XX_WHITE,	DataReceived,			"",		true);		//是否接收到数据

		int nConnected = -1;
		if (USE_Communicate)
		{
			#if (defined USE_WIFI) || (defined USE_BLE) || (defined USE_BT_SERIAL)
				nConnected = Connected?1:0;
			#endif
		}
		Display_Content(24, "Connected",		ST77XX_WHITE,	nConnected,				"",		true);		//是否连接成功
	}
	else
	{
		DisplayType = 0;
	}

	if (DisplayType == 0)
	{
		if (boolDisplay)
		{
			boolDisplay = false;
			digitalWrite(TFT_BLK_Pin, LOW);
			tft.fillScreen(ST77XX_BLACK);
		}
	}
	else
	{
		if (! boolDisplay)
		{
			boolDisplay = true;
			digitalWrite(TFT_BLK_Pin, HIGH);
		}
	}
}
//---------------------------------------------------------------------------
//获取电量百分比
int GetPowerPercent(const double fV)
{
	int nV = (int)trunc(fV * 1000);

	if      (nV >= 4200) return 100;
	else if (nV >= 4060) return (int)map(nV, 4060, 4200, 90, 100);
	else if (nV >= 3980) return (int)map(nV, 3980, 4060, 80,  90);
	else if (nV >= 3920) return (int)map(nV, 3920, 3980, 70,  80);
	else if (nV >= 3870) return (int)map(nV, 3870, 3920, 60,  70);
	else if (nV >= 3820) return (int)map(nV, 3820, 3870, 50,  60);
	else if (nV >= 3790) return (int)map(nV, 3790, 3820, 40,  50);
	else if (nV >= 3770) return (int)map(nV, 3770, 3790, 30,  40);
	else if (nV >= 3740) return (int)map(nV, 3740, 3770, 20,  30);
	else if (nV >= 3700) return (int)map(nV, 3700, 3740, 15,  20);
	else if (nV >= 3680) return (int)map(nV, 3680, 3700, 10,  15);
	else if (nV >= 3450) return (int)map(nV, 3450, 3680,  5,  10);
	else if (nV >= 3000) return (int)map(nV, 3000, 3450,  0,   5);
	else                 return 0;
}
//---------------------------------------------------------------------------
//获取供电电量百分比
bool GetPowerPercent(const int nLowPercent = 15)
{
	Power_Percent = GetPowerPercent((double)analogRead(Power_Pin) * 3.3 * 2 / Power_BatteryNum / 4095);
	return (Power_Percent >= nLowPercent);
}
//---------------------------------------------------------------------------
//获取姿态
bool GetAngle()
{
	if (! MPU6050.GetData()) return false;

//	if (MPU6050.Data.AngleX < 0) MPU6050.Data.AngleX += 180; else if (MPU6050.Data.AngleX > 0) MPU6050.Data.AngleX -= 180;
//	MPU6050.Data.GyroY = -MPU6050.Data.GyroY;
//	MPU6050.Data.GyroZ = -MPU6050.Data.GyroZ;

	return true;
}
//---------------------------------------------------------------------------
//获取joystick正反面
bool GetJoystickSide()
{
	return (abs(MPU6050.Data.AngleX) > 135);
}
//---------------------------------------------------------------------------
//发送控制反馈
void Ctrl_Send(const String strCtrlType, const float fCtrlValue, const uint32_t nDelayTime = 1)
{
	String strSend = strCtrlType + "/" + String(fCtrlValue);

	uint8_t nChecksum = 0;
	for (int nIndex = 0; nIndex < strSend.length(); nIndex++) nChecksum += (uint8_t)strSend[nIndex];
	strSend += (char)nChecksum;

	if (USE_Communicate)
	{
		#ifdef USE_WIFI
			WIFI_Send(strSend);
		#endif

		#ifdef USE_BLE
			BLE_Send(strSend);
		#endif

		#ifdef USE_BT_SERIAL
			BT_SERIAL_Send(strSend);
		#endif
	}
	else
	{
		#ifdef USE_WIRELESS_SERIAL
			WIRELESS_SERIAL_Send(strSend);
		#endif
	}

	if (nDelayTime > 0) vTaskDelay(nDelayTime);
}
//---------------------------------------------------------------------------
//处理控制内容
void Ctrl_Receive_Process(const String strCtrlType, const float fCtrlValue)
{
	if      (strCtrlType == "00") F_PIDTimeLoop				= fCtrlValue;	//实际PID计算周期（毫秒）
	else if (strCtrlType == "01") F_AngleX					= fCtrlValue;	//MPU6050姿态，X轴角度
	else if (strCtrlType == "02") F_AngleY					= fCtrlValue;	//MPU6050姿态，Y轴角度
	else if (strCtrlType == "03") F_AngleZ					= fCtrlValue;	//MPU6050姿态，Z轴角度
	else if (strCtrlType == "04") F_GyroX					= fCtrlValue;	//MPU6050姿态，X轴角速度
	else if (strCtrlType == "05") F_GyroY					= fCtrlValue;	//MPU6050姿态，Y轴角速度
	else if (strCtrlType == "06") F_GyroZ					= fCtrlValue;	//MPU6050姿态，Z轴角速度
	else if (strCtrlType == "07") F_AccX					= fCtrlValue;	//MPU6050姿态，X轴加速度
	else if (strCtrlType == "08") F_AccY					= fCtrlValue;	//MPU6050姿态，Y轴加速度
	else if (strCtrlType == "09") F_AccZ					= fCtrlValue;	//MPU6050姿态，Z轴加速度
	else if (strCtrlType == "10") F_SR04					= fCtrlValue;	//SR04超声波测距
	else if (strCtrlType == "11") F_PWM						= fCtrlValue;	//PWM
	else if (strCtrlType == "12") F_PWMTurn					= fCtrlValue;	//转向PWM
	else if (strCtrlType == "13") F_L_PWM					= fCtrlValue;	//左轮PWM
	else if (strCtrlType == "14") F_R_PWM					= fCtrlValue;	//右轮PWM
	else if (strCtrlType == "15") F_L_ENC_Count				= fCtrlValue;	//左轮编码器计数总计
	else if (strCtrlType == "16") F_R_ENC_Count				= fCtrlValue;	//右轮编码器计数总计
	else if (strCtrlType == "17") F_Power_Percent			= fCtrlValue;	//供电电量百分比
	else if (strCtrlType == "21") F_Balance					= fCtrlValue;	//平衡角度（°）
	else if (strCtrlType == "22") F_MoveSpeed				= fCtrlValue;	//移动速度（一个PID计算周期内左右编码器计数之和）
	else if (strCtrlType == "23") F_TurnSpeed				= fCtrlValue;	//转向角速度（°/s）
	else if (strCtrlType == "28") F_AutoPatrol				= fCtrlValue;	//是否自动巡逻
	else if (strCtrlType == "29") F_Feedback				= fCtrlValue;	//是否反馈数据
	else if (strCtrlType == "31") F_kap						= fCtrlValue;	//角度环kp
	else if (strCtrlType == "32") F_kai						= fCtrlValue;	//角度环ki
	else if (strCtrlType == "33") F_kad						= fCtrlValue;	//角度环kd
	else if (strCtrlType == "34") F_kaf						= fCtrlValue;	//角度环kf
	else if (strCtrlType == "35") F_kam						= fCtrlValue;	//角度环km
	else if (strCtrlType == "36") F_kan						= fCtrlValue;	//角度环kn
	else if (strCtrlType == "41") F_ksp						= fCtrlValue;	//速度环kp
	else if (strCtrlType == "42") F_ksi						= fCtrlValue;	//速度环ki
	else if (strCtrlType == "43") F_ksd						= fCtrlValue;	//速度环kd
	else if (strCtrlType == "44") F_ksf						= fCtrlValue;	//速度环kf
	else if (strCtrlType == "45") F_ksm						= fCtrlValue;	//速度环km
	else if (strCtrlType == "46") F_ksn						= fCtrlValue;	//速度环kn
	else if (strCtrlType == "51") F_ktp						= fCtrlValue;	//转向环kp
	else if (strCtrlType == "52") F_kti						= fCtrlValue;	//转向环ki
	else if (strCtrlType == "53") F_ktd						= fCtrlValue;	//转向环kd
	else if (strCtrlType == "54") F_ktf						= fCtrlValue;	//转向环kf
	else if (strCtrlType == "55") F_ktm						= fCtrlValue;	//转向环km
	else if (strCtrlType == "56") F_ktn						= fCtrlValue;	//转向环kn
	else if (strCtrlType == "81") F_EmergencyStopStatus		= fCtrlValue;	//急停状态
	else if (strCtrlType == "91") F_WIRELESS_SERIAL_Baud	= fCtrlValue;	//WIRELESS_SERIAL波特率
	else if (strCtrlType == "92") F_Feedback_Interval		= fCtrlValue;	//反馈数据间隔时间
	else if (strCtrlType == "93") F_Power_BatteryNum		= fCtrlValue;	//供电电池数量
	else if (strCtrlType == "F1") {WIRELESS_SERIAL_Baud		= fCtrlValue;	WriteParam(strCtrlType, fCtrlValue);}	//WIRELESS_SERIAL波特率
}
//---------------------------------------------------------------------------
//接收控制内容
void Ctrl_Receive()
{
	while (true)
	{
		static unsigned long nLastTime = millis();
		unsigned long now = millis();

		String strReceive = "";

		bool boolSerial = Serial.available();
		if (boolSerial) strReceive = Serial.readStringUntil('\n');

		if (USE_Communicate)
		{
			#ifdef USE_WIFI
				if (strReceive.isEmpty()) strReceive = WIFI_Receive();
			#endif

			#ifdef USE_BLE
				if (strReceive.isEmpty()) strReceive = BLE_Receive();
			#endif

			#ifdef USE_BT_SERIAL
				if (strReceive.isEmpty()) strReceive = BT_SERIAL_Receive();
			#endif
		}
		else
		{
			#ifdef USE_WIRELESS_SERIAL
				if (strReceive.isEmpty()) strReceive = WIRELESS_SERIAL_Receive();
			#endif
		}

		if (strReceive.isEmpty())
		{
			if (now - nLastTime >= 1000)
			{
				nLastTime = now;
				DataReceived = 0;
				Ctrl_Send("A9", 1);  //发送反馈数据编码
			}

			break;
		}
		else
		{
			if (strReceive.substring(2, 3) == "/")
			{
				if (boolSerial) strReceive += " ";
				uint8_t nChecksum = 0;
				for (int nIndex = 0; nIndex < strReceive.length() - 1; nIndex++) nChecksum += (uint8_t)strReceive[nIndex];

				if (boolSerial || (nChecksum == (uint8_t)strReceive[strReceive.length() - 1]))
				{
					nLastTime = now;
					DataReceived = 1;

					String strCtrlType = strReceive.substring(0, 2);
					float fCtrlValue = strReceive.substring(3, strReceive.length() - 1).toFloat();
					Ctrl_Receive_Process(strCtrlType, fCtrlValue);

					if (boolSerial)
					{
						Ctrl_Send(strCtrlType, fCtrlValue);
						Serial.print(strCtrlType);
						Serial.print(", ");
						Serial.print(fCtrlValue);
						Serial.println();
					}
				}
			}
		}
	}
}
//---------------------------------------------------------------------------
//处理控制
void Ctrl_Process()
{
	//判断自动巡逻状态是否同步
	if ((AutoPatrol < 0) || (AutoPatrol > 2)) AutoPatrol = 0;
	if (AutoPatrol != F_AutoPatrol) Ctrl_Send("A8", AutoPatrol);

	//判断是否处于自动巡逻状态
	if (AutoPatrol == 0)
	{
		int nMoveSpeed = 0;
		int nTurnSpeed = 0;

		if ((! GetJoystickSide()) && (digitalRead(EmeStop_Pin) == LOW))
		{
			int nMoveAngle = (int)MPU6050.Data.AngleY;
			int nTurnAngle = (int)MPU6050.Data.AngleX;

			if (abs(nMoveAngle) < MoveAngleMin) nMoveAngle = 0;
			if (abs(nMoveAngle) > MoveAngleMax) nMoveAngle = ((nMoveAngle > 0)?MoveAngleMax:-MoveAngleMax);

			if (abs(nTurnAngle) < TurnAngleMin) nTurnAngle = 0;
			if (abs(nTurnAngle) > TurnAngleMax) nTurnAngle = ((nTurnAngle > 0)?TurnAngleMax:-TurnAngleMax);

			nMoveSpeed = -map(nMoveAngle, -MoveAngleMax, MoveAngleMax, -MoveSpeedMax, MoveSpeedMax);
			nTurnSpeed =  map(nTurnAngle, -TurnAngleMax, TurnAngleMax, -TurnSpeedMax, TurnSpeedMax);
		}

		if (nMoveSpeed != F_MoveSpeed) Ctrl_Send("A2", nMoveSpeed);
		if (nTurnSpeed != F_TurnSpeed) Ctrl_Send("A3", nTurnSpeed);
	}
}
//---------------------------------------------------------------------------
//检测按钮
uint8_t ButtonClick()
{
	static bool boolBtnDown = false;
	static uint8_t nBtnClickNum = 0;

	static unsigned long nLastTime_BtnDown  = millis() - 1000;
	static unsigned long nLastTime_BtnClick = millis() - 1000;

	unsigned long now = millis();
	uint8_t nResult = 0;

	//判断按钮是否按下或抬起
	if (digitalRead(EmeStop_Pin) == (boolBtnDown?HIGH:LOW))
	{
		vTaskDelay(20);  //去抖
		if (digitalRead(EmeStop_Pin) == (boolBtnDown?HIGH:LOW))
		{
			boolBtnDown = (! boolBtnDown);

			if (boolBtnDown)
			{
				nLastTime_BtnDown = now;
			}
			else
			{
				//如果按钮从按下到抬起的时间小于300ms，则认为是点击
				if (now - nLastTime_BtnDown < 300)
				{
					//如果两次点击间隔时间小于300ms，则认为是多次点击
					if (now - nLastTime_BtnClick < 300) nBtnClickNum++; else nBtnClickNum = 1;
					nLastTime_BtnClick = now;
				}
				else
				{
					nBtnClickNum = 0;
				}
			}
		}
	}

	//如果已经是3次点击，或者上次点击时间是300ms之前，则返回点击次数
	if ((nBtnClickNum >= 3) || (now - nLastTime_BtnClick >= 300))
	{
		nResult = nBtnClickNum;
		nBtnClickNum = 0;
	}

	return nResult;
}
//---------------------------------------------------------------------------
