//---------------------------------------------------------------------------
#include "Global.h"
//---------------------------------------------------------------------------
//任务
void Task_Ctrl			(void* pvParameters);	//接收控制内容
void Task_Button		(void* pvParameters);	//检测按钮
void Task_KeepConnect	(void* pvParameters);	//保持数据通讯连接
void Task_Core0			(void* pvParameters);	//其他
//---------------------------------------------------------------------------
void setup()
{
	//初始化串口
	Serial.begin(115200);
	while (! Serial);
	Serial.println("\nprogram start");

	//关闭低电压检测，避免无限重启
//	WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);

	//初始化显示屏
	Display_Init();
	Display_Label(0, "initialization ...");

	//初始化急停引脚
	pinMode(EmeStop_Pin, INPUT_PULLUP);

	//初始化电压ADC引脚
	pinMode(Power_Pin, INPUT);

	//初始化无线透传引脚
	pinMode(Wireless_CS_Pin, OUTPUT);
	digitalWrite(Wireless_CS_Pin, HIGH);

	//初始化参数
	Display_Label(1, "EEPROM ...");
	if (EEPROM.begin(0x100 * sizeof(float) + 2100))
	{
		ReadAllParams();
		Display_Label(1, "EEPROM ok");
	}
	else
	{
		Display_Label(1, "EEPROM failure");
	}

	//初始化姿态模块
	Display_Label(2, "MPU6050 ...");
	Serial2.begin(115200, SERIAL_8N1, MPU6050_TX_Pin, MPU6050_RX_Pin);
	if (MPU6050.begin(&Serial2)) Display_Label(2, "MPU6050 ok"); else Display_Label(2, "MPU6050 failure");

	//初始化是否启用数据通讯功能（上电的时候，如果按住急停按钮，则启用数据通讯功能）
	USE_Communicate = (digitalRead(EmeStop_Pin) == LOW);

	//初始化WIFI
	#ifdef USE_WIFI
		if (USE_Communicate)
		{
			Display_Label(4, "WIFI ...");
			if (WIFI_Init())
			{
				//初始化TCP Client
				Display_Label(4, "TCP Client ...");
				if (TCP_Client_Init()) Display_Label(4, "TCP Client ok"); else Display_Label(4, "TCP Client failure");
			}
			else
			{
				Display_Label(4, "WIFI failure");
			}
		}
		else
		{
			Display_Label(4, "WIFI ignore");
		}
	#endif

	//初始化BLE
	#ifdef USE_BLE
		if (USE_Communicate)
		{
			Display_Label(4, "BLE ...");
			if (BLE_Init()) Display_Label(4, "BLE ok (" + String(BLE_Name) + ")"); else Display_Label(4, "BLE failure");
		}
		else
		{
			Display_Label(4, "BLE ignore");
		}
	#endif

	//初始化BT_SERIAL
	#ifdef USE_BT_SERIAL
		if (USE_Communicate)
		{
			Display_Label(4, "BT SERIAL ...");
			if (BT_SERIAL_Init()) Display_Label(4, "BT SERIAL ok (" + String(BT_SERIAL_Name) + ")"); else Display_Label(4, "BT SERIAL failure");
		}
		else
		{
			Display_Label(4, "BT SERIAL ignore");
		}
	#endif

	//初始化WIRELESS_SERIAL
	#ifdef USE_WIRELESS_SERIAL
		if (! USE_Communicate)
		{
			Display_Label(5, "WIRELESS SERIAL ...");
			digitalWrite(Wireless_CS_Pin, LOW);
			if (WIRELESS_SERIAL_Init(&Serial1, WIRELESS_SERIAL_Baud, Wireless_TX_Pin, Wireless_RX_Pin)) Display_Label(5, "WIRELESS SERIAL ok"); else Display_Label(5, "WIRELESS SERIAL failure");
		}
		else
		{
			Display_Label(5, "WIRELESS SERIAL ignore");
		}
	#endif

	//初始化任务
	Display_Label(6, "task ...");
	bool boolTask = true;
	boolTask = (boolTask && (xTaskCreatePinnedToCore(Task_Ctrl,			"Task_Ctrl",		2000, NULL, 1, NULL, 0) == pdPASS));
	boolTask = (boolTask && (xTaskCreatePinnedToCore(Task_Button,		"Task_Button",		2000, NULL, 1, NULL, 0) == pdPASS));
	boolTask = (boolTask && (xTaskCreatePinnedToCore(Task_KeepConnect,	"Task_KeepConnect",	3000, NULL, 1, NULL, 0) == pdPASS));
	boolTask = (boolTask && (xTaskCreatePinnedToCore(Task_Core0,		"Task_Core0",		3000, NULL, 1, NULL, 0) == pdPASS));
	if (boolTask) Display_Label(6, "task ok"); else Display_Label(6, "task failure");

	//初始化完成
	Display_Label(0, "initialization ok");
}
//---------------------------------------------------------------------------
void loop()
{
	static unsigned long nLastTime = millis();
	unsigned long now = millis();
	if (now - nLastTime >= CtrlTime)
	{
		nLastTime = now;

		//获取姿态
		GetAngle();

		//处理控制
		Ctrl_Process();
	}
}
//---------------------------------------------------------------------------
void Task_Ctrl(void* pvParameters)
{
	while (true)
	{
		//接收控制内容
		Ctrl_Receive();
		vTaskDelay(1);
	}
}
//---------------------------------------------------------------------------
void Task_Button(void* pvParameters)
{
	while (true)
	{
		//检测按钮
		switch (ButtonClick())
		{
			//单击
			case 1:
				break;

			//双击
			case 2:
				//切换显示内容
				if (GetJoystickSide()) DisplayType++;
				break;

			//三击
			case 3:
				//切换自动巡逻
				if (! GetJoystickSide()) AutoPatrol++;
				break;
		}

		vTaskDelay(1);
	}
}
//---------------------------------------------------------------------------
void Task_KeepConnect(void* pvParameters)
{
	while (true)
	{
		if (USE_Communicate)
		{
			#ifdef USE_WIFI
				WIFI_KeepConnect();
			#endif

			#ifdef USE_BLE
				BLE_KeepConnect();
			#endif

			#ifdef USE_BT_SERIAL
				BT_SERIAL_KeepConnect();
			#endif
		}

		vTaskDelay(1000);
	}
}
//---------------------------------------------------------------------------
void Task_Core0(void* pvParameters)
{
	unsigned long nStartTime = millis();

	while (true)
	{
		//获取供电电量百分比
		GetPowerPercent();

		//显示内容
		static bool boolDisplayContent = false;
		if (boolDisplayContent)
		{
			Display_Content();
		}
		else
		{
			boolDisplayContent = (millis() - nStartTime >= 5000);
//			if (boolDisplayContent) Display_Init();
		}

		//默认看门狗CONFIG_TASK_WDT_TIMEOUT_S=5秒，如果程序运行期间没有任何空闲时间，会导致系统无法自动喂狗，或者索性禁用看门狗disableCore0WDT();
		vTaskDelay(1);
	}
}
//---------------------------------------------------------------------------
