//---------------------------------------------------------------------------
#include "MPU6050.h"
//---------------------------------------------------------------------------
uint8_t TMPU6050::i2cWrite(const uint8_t nRegisterAddress, const uint8_t nData, const boolean boolSendStop)
{
	return i2cWrite(nRegisterAddress, &nData, 1, boolSendStop);  //成功则返回0
}
//---------------------------------------------------------------------------
uint8_t TMPU6050::i2cWrite(const uint8_t nRegisterAddress, const uint8_t* pnData, const uint8_t nLength, const boolean boolSendStop)
{
	FpI2CInterface->beginTransmission(FnI2C_Address);
	FpI2CInterface->write(nRegisterAddress);
	FpI2CInterface->write(pnData, nLength);
	uint8_t nReturnCode = FpI2CInterface->endTransmission(boolSendStop);  //成功则返回0

	if (nReturnCode)
	{
		Serial.print(F("i2cWrite failed: "));
		Serial.println(nReturnCode);
	}

	return nReturnCode;
}
//---------------------------------------------------------------------------
uint8_t TMPU6050::i2cRead(const uint8_t nRegisterAddress, uint8_t* pnData, const uint8_t nBytes)
{
	FpI2CInterface->beginTransmission(FnI2C_Address);
	FpI2CInterface->write(nRegisterAddress);
	uint8_t nReturnCode = FpI2CInterface->endTransmission(false);  //成功则返回0，先不释放i2c总线

	if (nReturnCode)
	{
		Serial.print(F("i2cRead failed: "));
		Serial.println(nReturnCode);
		return nReturnCode;
	}

	FpI2CInterface->requestFrom(FnI2C_Address, nBytes, (uint8_t)true);

	for (uint8_t i = 0; i < nBytes; i++)
	{
		if (FpI2CInterface->available())
		{
			pnData[i] = FpI2CInterface->read();
		}
		else
		{
			uint32_t nTimeOut = micros();
			while ((micros() - nTimeOut < FnI2C_Timeout) && (! FpI2CInterface->available()));

			if (FpI2CInterface->available())
			{
				pnData[i] = FpI2CInterface->read();
			}
			else
			{
				Serial.println(F("i2cRead timeout"));
				return 5;  //FpI2CInterface->endTransmission没有使用过这个返回值
			}
		}
	}

	return nReturnCode;
}
//---------------------------------------------------------------------------
uint8_t TMPU6050::i2cData()
{
	uint8_t nReturnCode;

	for (int nLoop = 0; nLoop < 3; nLoop++)
	{
		//获取I2C数据
		nReturnCode = i2cRead(0x3B, FnI2C_Buffer, 14);
		if (nReturnCode == 0)
		{
			//获取加速度计数据
			FdI2C_AccX  = (int16_t)((FnI2C_Buffer[0]  << 8) | FnI2C_Buffer[1]);  FdI2C_AccX  /= FdI2C_AccConvert;  FdI2C_AccX  += FdAccX_Err;
			FdI2C_AccY  = (int16_t)((FnI2C_Buffer[2]  << 8) | FnI2C_Buffer[3]);  FdI2C_AccY  /= FdI2C_AccConvert;  FdI2C_AccY  += FdAccY_Err;
			FdI2C_AccZ  = (int16_t)((FnI2C_Buffer[4]  << 8) | FnI2C_Buffer[5]);  FdI2C_AccZ  /= FdI2C_AccConvert;  FdI2C_AccZ  += FdAccZ_Err;

			//获取陀螺仪数据
			FdI2C_GyroX = (int16_t)((FnI2C_Buffer[8]  << 8) | FnI2C_Buffer[9]);  FdI2C_GyroX /= FdI2C_GyroConvert; FdI2C_GyroX += FdGyroX_Err;
			FdI2C_GyroY = (int16_t)((FnI2C_Buffer[10] << 8) | FnI2C_Buffer[11]); FdI2C_GyroY /= FdI2C_GyroConvert; FdI2C_GyroY += FdGyroY_Err;
			FdI2C_GyroZ = (int16_t)((FnI2C_Buffer[12] << 8) | FnI2C_Buffer[13]); FdI2C_GyroZ /= FdI2C_GyroConvert; FdI2C_GyroZ += FdGyroZ_Err;

			//获取温度数据
			int16_t nTemper = (int16_t)((FnI2C_Buffer[6] << 8) | FnI2C_Buffer[7]);
			Data.Temper = TTL_Calc_Tempe(nTemper);

			break;
		}
	}

	return nReturnCode;
}
//---------------------------------------------------------------------------
void TMPU6050::KalmanInit()
{
	//计算角度
	double dAngleX = atan2(FdI2C_AccY, FdI2C_AccZ) * RAD_TO_DEG;
	double dAngleY = atan(-FdI2C_AccX / sqrt(FdI2C_AccY * FdI2C_AccY + FdI2C_AccZ * FdI2C_AccZ)) * RAD_TO_DEG;

	//设定起始角度
	FKalmanX.setAngle(dAngleX);
	FKalmanY.setAngle(dAngleY);

	//设定起始时间
	FnKalmanTimer = micros();
}
//---------------------------------------------------------------------------
void TMPU6050::KalmanFilter()
{
	//计算时间变化
	double dDeltaTime = (double)(micros() - FnKalmanTimer) / 1000000.0;
	FnKalmanTimer = micros();

	//计算角度
	double dAngleX = atan2(FdI2C_AccY, FdI2C_AccZ) * RAD_TO_DEG;
	double dAngleY = atan(-FdI2C_AccX / sqrt(FdI2C_AccY * FdI2C_AccY + FdI2C_AccZ * FdI2C_AccZ)) * RAD_TO_DEG;

	//判断X轴角度是否在-180到180度之间翻转
	if (((dAngleX < -90) && (FdKalmanAngleX > 90)) || ((dAngleX > 90) && (FdKalmanAngleX < -90)))
	{
		FKalmanX.setAngle(dAngleX);
		FdKalmanAngleX = dAngleX;
	}
	else
	{
		FdKalmanAngleX = FKalmanX.getAngle(dAngleX, FdI2C_GyroX, dDeltaTime);
	}

	//如果X轴角度大于90度，则翻转Y轴角速度
	if (abs(FdKalmanAngleX) > 90) FdI2C_GyroY = -FdI2C_GyroY;
	FdKalmanAngleY = FKalmanY.getAngle(dAngleY, FdI2C_GyroY, dDeltaTime);

	//返回数据
	Data.AccX    = FdI2C_AccX;
	Data.AccY    = FdI2C_AccY;
	Data.AccZ    = FdI2C_AccZ;
	Data.GyroX   = FdI2C_GyroX;
	Data.GyroY   = FdI2C_GyroY;
	Data.GyroZ   = FdI2C_GyroZ;
	Data.AngleX  = FdKalmanAngleX;
	Data.AngleY  = FdKalmanAngleY;
	Data.AngleZ += FdI2C_GyroZ * dDeltaTime;	//= atan(FdI2C_AccZ / sqrt(FdI2C_AccX * FdI2C_AccX + FdI2C_AccY * FdI2C_AccY)) * RAD_TO_DEG;
}
//---------------------------------------------------------------------------
void TMPU6050::Init(const uint8_t nInterfaceType, const double dAccX_Err, const double dAccY_Err, const double dAccZ_Err, const double dGyroX_Err, const double dGyroY_Err, const double dGyroZ_Err, const double dAngleX_Err, const double dAngleY_Err, const double dAngleZ_Err)
{
	//初始化接口类型
	FnInterfaceType = nInterfaceType;

	//初始化传感器偏差
	FdAccX_Err   = dAccX_Err;   FdAccY_Err   = dAccY_Err;   FdAccZ_Err   = dAccZ_Err;
	FdGyroX_Err  = dGyroX_Err;  FdGyroY_Err  = dGyroY_Err;  FdGyroZ_Err  = dGyroZ_Err;
	FdAngleX_Err = dAngleX_Err; FdAngleY_Err = dAngleY_Err; FdAngleZ_Err = dAngleZ_Err;

	//初始化数据
	Data.AccX   = 0; Data.AccY   = 0; Data.AccZ   = 0;
	Data.GyroX  = 0; Data.GyroY  = 0; Data.GyroZ  = 0;
	Data.AngleX = 0; Data.AngleY = 0; Data.AngleZ = 0;
	Data.Temper = 0;
}
//---------------------------------------------------------------------------
boolean TMPU6050::begin(HardwareSerial* pSerialInterface, const double dAccX_Err, const double dAccY_Err, const double dAccZ_Err, const double dGyroX_Err, const double dGyroY_Err, const double dGyroZ_Err, const double dAngleX_Err, const double dAngleY_Err, const double dAngleZ_Err)
{
	//初始化数据
	FpSerialInterface = pSerialInterface;
	Init(0, dAccX_Err, dAccY_Err, dAccZ_Err, dGyroX_Err, dGyroY_Err, dGyroZ_Err, dAngleX_Err, dAngleY_Err, dAngleZ_Err);

	//初始化模块
//	FpSerialInterface->write(0xFF); FpSerialInterface->write(0xAA); FpSerialInterface->write(0x61);		//使用串口，禁用I2C。0x62：禁用串口，使用I2C
//	FpSerialInterface->write(0xFF); FpSerialInterface->write(0xAA); FpSerialInterface->write(0x63);		//设置速率，波特率115200，回传速率100Hz。0x64：9600，20Hz
//	FpSerialInterface->write(0xFF); FpSerialInterface->write(0xAA); FpSerialInterface->write(0x65);		//设置模块水平放置。0x66：垂直放置
//	FpSerialInterface->write(0xFF); FpSerialInterface->write(0xAA); FpSerialInterface->write(0x52);		//角度初始化，使Z轴角度归零
//	FpSerialInterface->write(0xFF); FpSerialInterface->write(0xAA); FpSerialInterface->write(0x67);		//加速度计校准，校准加速度零偏
//	FpSerialInterface->write(0xFF); FpSerialInterface->write(0xAA); FpSerialInterface->write(0x60);		//休眠及解休眠，待机模式和工作模式

	//等待传感器稳定
	delay(100);
	return true;
}
//---------------------------------------------------------------------------
boolean TMPU6050::begin(TwoWire* pI2CInterface, const double dAccX_Err, const double dAccY_Err, const double dAccZ_Err, const double dGyroX_Err, const double dGyroY_Err, const double dGyroZ_Err)
{
	//初始化数据
	FpI2CInterface = pI2CInterface;
	Init(1, dAccX_Err, dAccY_Err, dAccZ_Err, dGyroX_Err, dGyroY_Err, dGyroZ_Err);

	//获取WHO_AM_I，并校验
	if ((i2cRead(0x75, FnI2C_Buffer, 1) != 0) || (FnI2C_Buffer[0] != FnI2C_Address)) return false;

	//禁用睡眠模式，时钟源参考陀螺仪X轴
	if (i2cWrite(0x6B, 0x01, false) != 0) return false;

	//设置传感器参数
	FnI2C_Buffer[0] = 7;	//将采样频率设置为1kHz = 陀螺仪输出频率8kHz / (1 + 7)
	FnI2C_Buffer[1] = 0;	//禁用FSYNC（不采样连接到FSYNC引脚的外部信号）；加速度计滤波带宽为260Hz，延迟0毫秒，输出频率为1kHz；陀螺仪滤波带宽为256Hz，延迟0.98毫秒，输出频率为8kHz
	FnI2C_Buffer[2] = 24;	//设置陀螺仪量程，0：±250°/s、8：±500°/s、16：±1000°/s、24：±2000°/s，量程小则追踪快速动作，量程大则追踪慢速动作
	FnI2C_Buffer[3] = 24;	//设置加速度计量程，0：±2g、8：±4g、16：±8g、24：±16g

	//初始化加速度计、陀螺仪数据转换系数
	FdI2C_AccConvert  = pow(2, 16 - 1) / pow(2, FnI2C_Buffer[3] / 8 + 1);		//16位带符号，加速度计量程设置为±2g，16384 = 2^(16 - 1) / 2g
	FdI2C_GyroConvert = pow(2, 16 - 1) / pow(2, FnI2C_Buffer[2] / 8) / 250.0;	//16位带符号，陀螺仪量程设置为±250°/s，131.072 = 2^(16 - 1) / 250°/s

	//写入传感器参数
	if (i2cWrite(0x19, FnI2C_Buffer, 4, true) != 0) return false;

	//等待传感器稳定
	delay(100);

	//初始化卡尔曼滤波函数
	if (i2cData() != 0) return false;
	KalmanInit();
	return true;
}
//---------------------------------------------------------------------------
boolean TMPU6050::GetData(const boolean boolReversalX, const boolean boolReversalY, const boolean boolReversalZ)
{
	boolean boolGetData = false;

	//串口
	if (FnInterfaceType == 0)
	{
		boolean boolGetData_Acc = false, boolGetData_Gyro = false, boolGetData_Angle = false;

		while (FpSerialInterface->available())
    	{
			//如果是0x55，则认为是数据包头，然后读取后续的10个字节的数据
			FnTTL_Buffer[0] = FpSerialInterface->read();
			if (FnTTL_Buffer[0] == 0x55) FpSerialInterface->readBytes(&FnTTL_Buffer[1], 10); else continue;

			//校验和
			uint8_t nChecksum = FnTTL_Buffer[0] + FnTTL_Buffer[1] + FnTTL_Buffer[2] + FnTTL_Buffer[3] + FnTTL_Buffer[4] + FnTTL_Buffer[5] + FnTTL_Buffer[6] + FnTTL_Buffer[7] + FnTTL_Buffer[8] + FnTTL_Buffer[9];
			if (nChecksum == FnTTL_Buffer[10]) memcpy(&FsTTL_Data, FnTTL_Buffer, 11); else continue;

			//判断数据包类型
			switch (FsTTL_Data.Type)
			{
        		//加速度
            	case 0x51:
					Data.AccX   = TTL_Calc_Acc(FsTTL_Data.X); Data.AccX += FdAccX_Err; if (boolReversalX) Data.AccX = -Data.AccX;
					Data.AccY   = TTL_Calc_Acc(FsTTL_Data.Y); Data.AccY += FdAccY_Err; if (boolReversalY) Data.AccY = -Data.AccY;
					Data.AccZ   = TTL_Calc_Acc(FsTTL_Data.Z); Data.AccZ += FdAccZ_Err; if (boolReversalZ) Data.AccZ = -Data.AccZ;
        		    Data.Temper = TTL_Calc_Tempe(FsTTL_Data.T);
					boolGetData_Acc = true;
                	break;

		        //角速度
    	        case 0x52:
    		        Data.GyroX  = TTL_Calc_Gyro(FsTTL_Data.X); Data.GyroX += FdGyroX_Err; if (boolReversalX) Data.GyroX = -Data.GyroX;
            		Data.GyroY  = TTL_Calc_Gyro(FsTTL_Data.Y); Data.GyroY += FdGyroY_Err; if (boolReversalY) Data.GyroY = -Data.GyroY;
                	Data.GyroZ  = TTL_Calc_Gyro(FsTTL_Data.Z); Data.GyroZ += FdGyroZ_Err; if (boolReversalZ) Data.GyroZ = -Data.GyroZ;
	            	Data.Temper = TTL_Calc_Tempe(FsTTL_Data.T);
					boolGetData_Gyro = true;
    	            break;

		        //角度
    		    case 0x53:
        	        Data.AngleX = TTL_Calc_Angle(FsTTL_Data.X); Data.AngleX += FdAngleX_Err; if (boolReversalX) Data.AngleX = -Data.AngleX;
        		    Data.AngleY = TTL_Calc_Angle(FsTTL_Data.Y); Data.AngleY += FdAngleY_Err; if (boolReversalY) Data.AngleY = -Data.AngleY;
                	Data.AngleZ = TTL_Calc_Angle(FsTTL_Data.Z); Data.AngleZ += FdAngleZ_Err; if (boolReversalZ) Data.AngleZ = -Data.AngleZ;
		            Data.Temper = TTL_Calc_Tempe(FsTTL_Data.T);
					boolGetData_Angle = true;
    	            break;
			}
		}

		boolGetData = (boolGetData_Acc && boolGetData_Gyro && boolGetData_Angle);
	}
	//I2C
	else if (FnInterfaceType == 1)
	{
		//卡尔曼滤波函数
		if (i2cData() == 0)
		{
			KalmanFilter();

			//判断是否需要翻转数据
			if (boolReversalX) {Data.AccX = -Data.AccX; Data.GyroX = -Data.GyroX; Data.AngleX = -Data.AngleX;}
			if (boolReversalY) {Data.AccY = -Data.AccY; Data.GyroY = -Data.GyroY; Data.AngleY = -Data.AngleY;}
			if (boolReversalZ) {Data.AccZ = -Data.AccZ; Data.GyroZ = -Data.GyroZ; Data.AngleZ = -Data.AngleZ;}

			boolGetData = true;
		}
	}

	return boolGetData;
}
//---------------------------------------------------------------------------
